# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
from abc import abstractmethod
from typing import Callable, Dict, List, Optional, Tuple, Union

import torch
import torch.nn.functional as F
from nemo.utils import logging

# TODO: Move the cache implementation out of the Module class, and pass it as part of the forward so we can reset
# as needed in the inference pipeline.


class ConvolutionLayer(torch.nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int = 1,
        stride: int = 1,
        padding: Optional[int] = None,
        dilation: int = 1,
        bias: bool = True,
        is_causal: bool = False,
    ):
        """
        A convolutional layer that supports causal convolutions with padding. Replaces the standard MLP layer used in
        the original transformer.

        Args:
            in_channels (int): Number of input channels.
            out_channels (int): Number of output channels.
            kernel_size (int): Size of the convolving kernel.
            stride (int): Stride of the convolution.
            padding (Optional[int]): Padding added to both sides of the input. If None, it's calculated automatically.
            dilation (int): Spacing between kernel elements.
            bias (bool): If True, adds a learnable bias to the output.
            is_causal (bool): If True, uses causal convolution.
        """
        super().__init__()

        # Setup up padding; should be 0 if set to causal
        # If not causal and padding is None, set an appropriate value for padding
        self.causal_padding = None
        if is_causal:
            self.causal_padding = ((kernel_size - 1) * dilation, 0)
            if padding is not None:
                logging.warning(
                    f'{self} was initialized with is_causal set to True, and padding set to {padding}. '
                    f'The provided padding value will be ignored and set to {self.causal_padding}.'
                )
            padding = 0
        elif padding is None:
            if kernel_size % 2 == 0:
                raise ValueError("`kernel_size` must be odd when `padding` is None.")
            else:
                padding = int(dilation * (kernel_size - 1) / 2)

        self.is_causal = is_causal
        self.kernel_size = kernel_size
        self.dilation = dilation

        self.conv = torch.nn.Conv1d(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            bias=bias,
        )

    def forward(self, signal, signal_mask):
        # signal: (B, C, T)
        # signal_mask: (B, T)
        signal = signal * signal_mask.unsqueeze(1)
        if self.is_causal:  # TODO: maybe replace with identify rather than keep conditional if in forward
            signal = F.pad(signal, self.causal_padding)

        conv_signal = self.conv(signal)
        conv_signal = conv_signal * signal_mask.unsqueeze(1)

        return conv_signal


class PositionwiseConvFF(torch.nn.Module):
    def __init__(
        self,
        d_model: int,
        d_ffn: int,
        p_dropout: float,
        kernel_size: int = 1,
        bias: bool = False,
        is_causal: bool = True,
        non_linearity: Callable = torch.nn.GELU(approximate="tanh"),
    ):
        """
        Positionwise Convolutional Feed-Forward layer to replace the MLP layer in transformers.

        Module will take the input with d_model hidden state, project it to d_ffn hidden dimension, perform nonlinear
        transformation, and project the state back into d_model hidden dimension. Finally, it applied dropout.

        Args:
            d_model (int): Input and output dimension of the model.
            d_ffn (int): Hidden dimension of the feed-forward network (usually 4 * d_model).
            p_dropout (float): Dropout probability.
            kernel_size (int): Size of the convolving kernel.
            bias (bool): If True, adds a learnable bias to the convolution layers.
            is_causal (bool): If True, uses causal convolution.
            non_linearity (Callable): Activation function to use (default: GELU).
        """
        super().__init__()
        # d_ffn is usually 4*d_model
        self.d_model = d_model
        self.non_linearity = non_linearity

        self.proj = ConvolutionLayer(d_model, d_ffn, bias=bias, kernel_size=kernel_size, is_causal=is_causal)
        self.o_net = ConvolutionLayer(d_ffn, d_model, bias=bias, kernel_size=kernel_size, is_causal=is_causal)
        self.dropout = torch.nn.Dropout(p_dropout)

    def forward(self, x, x_mask):
        """
        x (B, T, C)
        x_mask (B, T)
        """
        x = self.non_linearity(self.proj(x.transpose(1, 2), x_mask))
        x = self.dropout(self.o_net(x, x_mask).transpose(1, 2))
        return x


class Attention(torch.nn.Module):
    def __init__(
        self,
        n_heads: int,
        d_model: int,
        p_dropout: float,
        is_causal: bool = True,
        d_head: Optional[int] = None,
    ):
        """
        Base Attention parent class. Users should not be instantiating this class, but rather use SelfAttention or
        CrossAttention classes as appropriate.
        Does DotProductionAttention and additionally dropout inside the module. The class does not currently support
        RoPE nor ALiBi.

        Args:
            n_heads (int): Number of attention heads.
            d_model (int): Dimension of the model.
            p_dropout (float): Dropout probability.
            is_causal (bool): Whether to use causal attention. Only supported when used in SelfAttention.
            d_head (int): Head dimension. Defaults to d_model // n_heads.
        """
        super().__init__()
        assert d_model % n_heads == 0, "d_model % n_head != 0"
        self.d_head = d_head if d_head is not None else d_model // n_heads
        self.n_heads = n_heads
        self.d_model = d_model
        self.scale = self.d_head**-0.5
        self.is_causal = is_causal
        self.o_net = torch.nn.Linear(n_heads * self.d_head, d_model, bias=False)
        self.dropout = torch.nn.Dropout(p_dropout)
        self.use_cache = False
        self.cache = self._init_cache()

    @abstractmethod
    def compute_qkv_and_mask(
        self,
        query: torch.Tensor,
        query_mask: Optional[torch.Tensor] = None,
        memory: Optional[torch.Tensor] = None,
        memory_mask: Optional[torch.Tensor] = None,
    ):
        pass

    @staticmethod
    def _init_cache() -> Dict[str, Optional[Union[bool, torch.Tensor]]]:
        return {
            'is_initialized': False,
            'self_k': None,
            'self_v': None,
            'cross_kv': None,
            'cross_k': None,
            'cross_v': None,
        }

    def reset_cache(self, use_cache: bool = False):
        self.use_cache = use_cache
        self.cache = self._init_cache()

    def attn_naive(
        self,
        query: torch.Tensor,
        query_mask: Optional[torch.Tensor] = None,
        memory: Optional[torch.Tensor] = None,
        memory_mask: Optional[torch.Tensor] = None,
        attn_prior: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, List[torch.Tensor]]:
        if self.use_cache:
            if self.cache['is_initialized']:
                query = query[:, -1:, :]
                query_mask = query_mask[:, -1:] if query_mask is not None else None
            else:
                self.cache['is_initialized'] = True

        # Calls into children classes to compute qkv tensors and mask tensor
        q, k, v, mask = self.compute_qkv_and_mask(
            query=query, query_mask=query_mask, memory=memory, memory_mask=memory_mask
        )

        # (B, T, nh, dh) -> (B, nh, T, dh)
        q = q.transpose(1, 2)
        k = k.transpose(1, 2)
        v = v.transpose(1, 2)

        B, T, _ = query.shape
        attn_score = torch.matmul(q, k.transpose(2, 3)) * self.scale
        if mask is not None:
            # assumes there's at least one mask
            attn_score.masked_fill_(mask == 0, float('-inf'))
        if self.is_causal:
            attn_score.masked_fill_(self.causal_mask[..., :T, :T] == 0, float('-inf'))

        # attn_prior or square mask or vanilla attention
        if attn_prior is not None:
            eps = torch.finfo(attn_prior.dtype).tiny
            attn_prior = attn_prior[:, :T]  # trim for inference
            attn_prior = attn_prior[:, None] + eps
            # Use PyTorch's built-in training flag to branch behavior
            if self.training:
                attn_prior_log = torch.log(attn_prior)
                attn_score_log = F.log_softmax(attn_score, dim=-1) + attn_prior_log
                if self.make_prior_window_strict:
                    # Make sure attention scores are lowest (eps) where prior is zero.
                    min_score = torch.log(torch.tensor(eps)).to(attn_score_log.device)
                    attn_score_log = attn_score_log.masked_fill(
                        attn_prior == 0, min_score
                    )  # Wherever prior is zero, set scores to eps.
                    attn_score_log = torch.clamp(
                        attn_score_log, min=min_score
                    )  # Make sure scores are not less than eps.
                attn_prob = F.softmax(attn_score_log, dim=-1)
            else:
                attn_prob = F.softmax(attn_score, dim=-1)
                attn_prob = attn_prob * attn_prior
                attn_prob = attn_prob / (attn_prob.sum(dim=-1, keepdim=True))  # normalize
        else:
            attn_prob = F.softmax(attn_score, dim=-1)

        if mask is not None:
            attn_prob = attn_prob.masked_fill(mask == 0, 0.0)
        attn_prob = self.dropout(attn_prob)

        y = torch.matmul(attn_prob, v)
        y = y.transpose(1, 2).contiguous().view(B, T, -1)

        return y, [attn_prob, attn_score]

    def forward(
        self,
        query: torch.Tensor,
        query_mask: Optional[torch.Tensor] = None,
        memory: Optional[torch.Tensor] = None,
        memory_mask: Optional[torch.Tensor] = None,
        attn_prior: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, List[torch.Tensor]]:
        """
        Forward pass of the Attention module.

        Args:
            query (torch.Tensor): Input tensor of shape (B, T1, C).
            query_mask (Optional[torch.Tensor]): Mask for query tensor of shape (B, T1).
            memory (Optional[torch.Tensor]): Memory tensor for cross-attention of shape (B, T2, C).
            memory_mask (Optional[torch.Tensor]): Mask for memory tensor of shape (B, T2).
            attn_prior (Optional[torch.Tensor]): Prior attention weights of shape (B, T1, T2).

        Returns:
            Tuple[torch.Tensor, List[torch.Tensor]]:
                - y: Attention module tensor output of shape (B, T1, C).
                - attn_prob: List containing attention probabilities and scores. returned only in attn_naive.
                    [0]: Attention probabilities used for logging during validation.
                    [1]: Attention scores used for CTC loss (only in naive attention).
        """

        y, attn_prob = self.attn_naive(query, query_mask, memory, memory_mask, attn_prior)
        y = self.dropout(self.o_net(y))

        return y, attn_prob


class SelfAttention(Attention):
    def __init__(
        self,
        n_heads: int,
        d_model: int,
        p_dropout: float,
        is_causal: bool = True,
        max_length_causal_mask: int = 4096,
    ):
        """
        Implements SelfAttention. See parent class for forward implementation.

        Args:
            n_heads (int): Number of attention heads.
            d_model (int): Dimension of the model.
            p_dropout (float): Dropout probability.
            is_causal (bool): Whether to use causal attention. Only supported when used in SelfAttention.
            max_length_causal_mask (int): Maximum sequence length for Attention module.
        """
        super().__init__(
            n_heads=n_heads,
            d_model=d_model,
            p_dropout=p_dropout,
            is_causal=is_causal,
        )
        if is_causal:
            if max_length_causal_mask is None or max_length_causal_mask < 0:
                raise ValueError(
                    "Self Attention was called with is_causal True, but received an inappropriate value"
                    f"of {max_length_causal_mask} for max_length_causal_mask"
                )
            self.register_buffer(
                "causal_mask",
                torch.tril(torch.ones(max_length_causal_mask, max_length_causal_mask)).view(
                    1, 1, max_length_causal_mask, max_length_causal_mask
                ),
            )
        self.qkv_net = torch.nn.Linear(d_model, 3 * n_heads * self.d_head, bias=False)

    def compute_qkv_and_mask(
        self,
        query: torch.Tensor,
        query_mask: Optional[torch.Tensor] = None,
        memory: Optional[torch.Tensor] = None,
        memory_mask: Optional[torch.Tensor] = None,
    ):
        B, T, _ = query.shape
        qkv = self.qkv_net(query).reshape(B, T, 3, self.n_heads, self.d_head)
        q, k, v = qkv.chunk(3, dim=2)
        q, k, v = q.squeeze(2), k.squeeze(2), v.squeeze(2)
        if self.use_cache:
            if self.cache['self_k'] is not None:
                k = torch.cat([self.cache['self_k'], k], dim=1)
                v = torch.cat([self.cache['self_v'], v], dim=1)
            self.cache['self_k'] = k
            self.cache['self_v'] = v

        mask = None
        if query_mask is not None:
            # query_mask is a boolean mask of shape (B, T)
            # mask should be of shape (B, 1, T, T) where mask[:,0,i,:] == mask[:,0,:,i] == query_mask
            mask = query_mask.unsqueeze(1) * query_mask.unsqueeze(2)
            mask = mask.unsqueeze(1)

        return q, k, v, mask


class CrossAttention(Attention):
    def __init__(
        self,
        n_heads: int,
        d_model: int,
        d_memory: int,
        p_dropout: float,
        make_prior_window_strict: bool = False,
        d_head: Optional[int] = None,
    ):
        """
        Implements CrossAttention. See parent class for forward implementation. Must be non-causal.

        Args:
            n_heads (int): Number of attention heads.
            d_model (int): Dimension of the model.
            d_memory (int): Dimension of the conditioning / cross-attention input.
            p_dropout (float): Dropout probability.
            make_prior_window_strict (bool): Make attention scores lowest where prior is zero.
            d_head (int): Head dimension. if None, defaults to d_model // n_heads in parent class.
        """
        super().__init__(
            n_heads=n_heads,
            d_model=d_model,
            p_dropout=p_dropout,
            is_causal=False,
            d_head=d_head,
        )
        self.q_net = torch.nn.Linear(d_model, n_heads * self.d_head, bias=False)
        self.kv_net = torch.nn.Linear(d_memory, 2 * n_heads * self.d_head, bias=False)
        self.make_prior_window_strict = make_prior_window_strict

    def compute_qkv_and_mask(
        self,
        query: torch.Tensor,
        query_mask: Optional[torch.Tensor] = None,
        memory: Optional[torch.Tensor] = None,
        memory_mask: Optional[torch.Tensor] = None,
    ):
        Bq, Tq, _ = query.shape
        Bkv, Tkv, _ = memory.shape
        q = self.q_net(query).reshape(Bq, Tq, self.n_heads, self.d_head)
        if self.use_cache and self.cache['cross_kv'] is not None:
            kv = self.cache['cross_kv']
        else:
            kv = self.kv_net(memory).reshape(Bkv, Tkv, 2, self.n_heads, self.d_head)

        if self.use_cache and self.cache['cross_k'] is not None:
            k = self.cache['cross_k']
            v = self.cache['cross_v']
        else:
            k, v = kv.chunk(2, dim=2)
            k, v = k.squeeze(2), v.squeeze(2)
            if self.use_cache:
                self.cache['cross_kv'] = kv
                self.cache['cross_k'] = k
                self.cache['cross_v'] = v

        mask = memory_mask[:, None, None] if memory_mask is not None else None
        return q, k, v, mask


class TransformerLayer(torch.nn.Module):
    def __init__(
        self,
        d_model: int,
        d_ffn: int,
        sa_n_heads: int,
        kernel_size: int,
        p_dropout: float,
        has_xattn: bool,
        xa_d_memory: Optional[int] = None,
        xa_n_heads: Optional[int] = None,
        xa_d_head: Optional[int] = None,
        is_causal: bool = True,
        apply_norm_to_cond: bool = True,
        max_length_causal_mask: int = 4096,
        conv_non_linearity: Callable = torch.nn.GELU(approximate="tanh"),
        make_prior_window_strict: bool = False,
    ):
        """
        One layer of the Transformer.
        Args:
            d_model <int>: Model dimension
            d_ffn <int>: Feed forward dimension (usually 4*d_model)
            sa_n_heads <int>: Number of attention heads used in self-attention
            kernel_size <int>: Convolution kernel size for FFN
            p_dropout <float>: Dropout probability
            has_xattn <bool>: Whether to use cross attention
            xa_d_memory <int>: Hidden dimension for cross attention
            xa_n_heads <int>: Number of attention heads used in cross attention
            xa_d_head <int>: Head dimension for cross attention. if None, defaults to d_model // xa_n_heads in Attention class.
            is_causal <bool>: Whether to use causal attention
            apply_norm_to_cond <bool>: Whether to apply normalization to conditioning tensor
            max_length_causal_mask <int>: Maximum length of causal mask
            conv_non_linearity <Callable>: Convolution non-linearity
            make_prior_window_strict <bool>: Make attention scores lowest where prior is zero.
        """
        super().__init__()
        self.has_xattn = has_xattn

        self.norm_self = torch.nn.LayerNorm(d_model, bias=False)
        self.self_attention = SelfAttention(
            n_heads=sa_n_heads,
            d_model=d_model,
            p_dropout=p_dropout,
            max_length_causal_mask=max_length_causal_mask,
            is_causal=is_causal,
        )

        if self.has_xattn:
            self.norm_xattn_query = torch.nn.LayerNorm(d_model, bias=False)
            self.cross_attention = CrossAttention(
                n_heads=xa_n_heads,
                d_model=d_model,
                d_memory=xa_d_memory,
                p_dropout=p_dropout,
                make_prior_window_strict=make_prior_window_strict,
                d_head=xa_d_head,
            )

            self.norm_xattn_memory = torch.nn.Identity()
            if apply_norm_to_cond:
                self.norm_xattn_memory = torch.nn.LayerNorm(xa_d_memory, bias=False)

        self.norm_pos_ff = torch.nn.LayerNorm(d_model, bias=False)
        self.pos_ff = PositionwiseConvFF(
            d_model, d_ffn, p_dropout, kernel_size=kernel_size, is_causal=is_causal, non_linearity=conv_non_linearity
        )

        self.use_cache = False
        self.cache = self._init_cache()

    @staticmethod
    def _init_cache() -> Dict:
        return {
            'self_attn_output': None,
            'cross_attn_output': None,
            'memory': None,
        }

    def reset_cache(self, use_cache=False):
        self.use_cache = use_cache
        self.cache = self._init_cache()
        self.self_attention.reset_cache(use_cache)
        if self.has_xattn:
            self.cross_attention.reset_cache(use_cache)

    def forward(
        self,
        x: torch.Tensor,
        x_mask: torch.Tensor,
        cond: Optional[torch.Tensor] = None,
        cond_mask: Optional[torch.Tensor] = None,
        attn_prior: Optional[torch.Tensor] = None,
    ) -> Dict:
        """
        Args:
            x <torch tensor> (B, T1, C): Input tensor
            x_mask <bool mask> (B, T1): Multiplicative mask where True means we keep the input, False we zero it out.
                Mask for self attention input.
            cond <torch tensor> (B, T2, C): Conditioning tensor
            cond_mask <bool mask> (B, T2): Multiplicative mask where True means we keep the input, False we zero
                it out. Mask for cross attention input if it exists.

        Returns dict with keys
            output <torch tensor> (B, T1, C): Output tensor
            attn_probabilities <dict>: Attention probabilities
        """
        x = x * x_mask.unsqueeze(-1)
        x_, s_attn_prob = self.self_attention(query=self.norm_self(x), query_mask=x_mask)
        if self.use_cache:
            if self.cache['self_attn_output'] is not None:
                x_ = torch.cat([self.cache['self_attn_output'], x_], dim=1)
            self.cache['self_attn_output'] = x_
        x = x + x_

        x_attn_prob = None
        if self.has_xattn and cond is not None:
            x_normed = self.norm_xattn_query(x)
            if self.use_cache and self.cache['memory'] is not None:
                memory = self.cache['memory']
            else:
                memory = self.norm_xattn_memory(cond)
                if self.use_cache:
                    self.cache['memory'] = memory

            x_res, x_attn_prob = self.cross_attention(
                query=x_normed, query_mask=x_mask, memory=memory, memory_mask=cond_mask, attn_prior=attn_prior
            )
            if self.use_cache:
                if self.cache['cross_attn_output'] is not None:
                    x_res = torch.cat([self.cache['cross_attn_output'], x_res], dim=1)
                self.cache['cross_attn_output'] = x_res
            x = x + x_res

        # mlp final projection
        x = x + self.pos_ff(self.norm_pos_ff(x), x_mask)
        x = x * x_mask.unsqueeze(-1)

        return {
            'output': x,
            'attn_probabilities': {'self_attn_probabilities': s_attn_prob, 'cross_attn_probabilities': x_attn_prob},
        }


class Transformer(torch.nn.Module):
    def __init__(
        self,
        n_layers: int,
        d_model: int,
        d_ffn: int,
        sa_n_heads: int,
        kernel_size: int,
        p_dropout: float = 0.0,
        p_dropout_out: float = 0.0,
        has_xattn: bool = False,
        xa_d_memory: Optional[int] = None,
        xa_n_heads: Optional[int] = None,
        xa_d_head: Optional[int] = None,
        is_causal: bool = True,
        apply_norm_to_cond: bool = True,
        apply_norm_out: bool = False,
        max_length_causal_mask: int = 4096,
        use_learnable_pos_emb: bool = False,
        conv_non_linearity: Callable = torch.nn.GELU(approximate="tanh"),
        make_prior_window_strict: bool = False,
    ):
        """
        Initializes a stack of transformer layers. Can be used for both encoder and decoder.
        Set is_causal is True for autoregressive models. Equivalent to TransformerBlock from Megatron-LM
        Args:
            n_layers <int>: Number of transformer layers
            d_model <int>: Model dimension
            d_ffn <int>: Feed forward dimension (usually 4*d_model)
            sa_n_heads <int>: Number of attention heads used in self-attention
            kernel_size <int>: Convolution kernel size for FFN
            p_dropout <float>: Dropout probability
            p_dropout_out <float>: Dropout probability for output
            has_xattn <bool>: Whether to use cross attention
            xa_d_memory <int>: Hidden dimension for cross attention; required if has_xattn is True
            xa_n_heads <int>: Number of attention heads used in cross attention; required if has_xattn is True
            xa_d_head <int>: Head dimension for cross attention. if None, defaults to d_model // xa_n_heads in Attention class.
            is_causal <bool>: Whether to make attention and the convolution feedforward networks causal.
            apply_norm_to_cond <bool>: Whether to apply normalization to conditioning tensor; conditioning tensor being
                the input to the memory part of cross-attention.
            apply_norm_out <bool>: Whether to apply normalization to output
            max_length_causal_mask <int>: Maximum length of causal mask
            use_learnable_pos_emb <bool>: Whether to add a learnable positionable embedding inside the class
            conv_non_linearity <Callable>: Convolution non-linearity
            make_prior_window_strict <bool>: Make attention scores lowest where prior is zero
        """
        if has_xattn and (xa_d_memory is None or xa_n_heads is None):
            raise ValueError("It requires that `xa_d_memory` and `xa_n_heads` are specified when `has_xattn` is True!")

        super().__init__()
        self.n_layers = n_layers
        self.dropout = torch.nn.Dropout(p_dropout)
        self.p_dropout_out = p_dropout_out

        self.dropout_out = torch.nn.Identity()
        if self.p_dropout_out > 0.0:
            self.dropout_out = torch.nn.Dropout(self.p_dropout_out)

        self.norm_out = torch.nn.Identity()
        if apply_norm_out:
            self.norm_out = torch.nn.LayerNorm(d_model, bias=False)

        self.layers = torch.nn.ModuleList()
        for _ in range(self.n_layers):
            self.layers.append(
                TransformerLayer(
                    d_model=d_model,
                    d_ffn=d_ffn,
                    sa_n_heads=sa_n_heads,
                    kernel_size=kernel_size,
                    p_dropout=p_dropout,
                    has_xattn=has_xattn,
                    xa_d_memory=xa_d_memory,
                    xa_n_heads=xa_n_heads,
                    xa_d_head=xa_d_head,
                    is_causal=is_causal,
                    apply_norm_to_cond=apply_norm_to_cond,
                    max_length_causal_mask=max_length_causal_mask,
                    conv_non_linearity=conv_non_linearity,
                    make_prior_window_strict=make_prior_window_strict,
                )
            )

        self.use_learnable_pos_emb = use_learnable_pos_emb
        self.position_embeddings = None
        if self.use_learnable_pos_emb:
            self.position_embeddings = torch.nn.Embedding(max_length_causal_mask, d_model)
        # Apply random uniform init for all layers, except for output layers: The second of the two layers in the MLP
        # and the last linear projection in dot product attention. The output layers are scaled depending on the
        # number of layers
        self.apply(self._init_weights_gpt2)
        for name, param in self.named_parameters():
            if 'o_net' in name and name.endswith('weight'):
                torch.nn.init.normal_(param, mean=0.0, std=0.02 / math.sqrt(2 * self.n_layers))

    def reset_cache(self, use_cache=False):
        for layer in self.layers:
            layer.reset_cache(use_cache)

    @staticmethod
    def _init_weights_gpt2(module):
        if isinstance(module, (torch.nn.Linear, torch.nn.Embedding, torch.nn.Conv1d)):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)

        if isinstance(module, torch.nn.Linear) and module.bias is not None:
            torch.nn.init.zeros_(module.bias)

    @staticmethod
    def _get_layer_inputs(
        idx: int,
        cond: Optional[Union[torch.Tensor, List[torch.Tensor]]],
        cond_mask: Optional[Union[torch.Tensor, List[torch.Tensor]]],
        attn_prior: Optional[Union[torch.Tensor, List[torch.Tensor]]],
        multi_encoder_mapping: Optional[List[Optional[int]]],
    ) -> Tuple[Optional[torch.Tensor], Optional[torch.Tensor], Optional[torch.Tensor]]:
        if multi_encoder_mapping is not None:
            if multi_encoder_mapping[idx] is None:
                return None, None, None
            else:
                _attn_prior = attn_prior[multi_encoder_mapping[idx]] if attn_prior is not None else None
                if isinstance(_attn_prior, list):
                    # @pneekhara: This means, we are passing layerwise attn_prior
                    _attn_prior = _attn_prior[idx]
                return (
                    cond[multi_encoder_mapping[idx]],
                    cond_mask[multi_encoder_mapping[idx]] if cond_mask is not None else None,
                    _attn_prior,
                )
        else:
            if isinstance(attn_prior, list):
                # @pneekhara: This means, we are passing layerwise attn_prior
                _attn_prior = attn_prior[idx]
            else:
                _attn_prior = attn_prior
            return cond, cond_mask, _attn_prior

    def forward(
        self,
        x: torch.Tensor,
        x_mask: torch.Tensor,
        cond: Optional[Union[torch.Tensor, List[torch.Tensor]]] = None,
        cond_mask: Optional[Union[torch.Tensor, List[torch.Tensor]]] = None,
        attn_prior: Optional[Union[torch.Tensor, List[torch.Tensor]]] = None,
        multi_encoder_mapping: Optional[List[Optional[int]]] = None,
        max_layer_idx: Optional[int] = None,
    ) -> Dict[str, Union[torch.Tensor, List]]:
        """
        Args:
            x <torch tensor> (B, T1, C):
            x_mask <bool mask> (B, T1): Multiplicative mask where True means we keep the input, False we zero it out.
                Mostly used in non-causal self-attention to zero out padding values. In causal self-attention, the
                causal mask will be used in place of this.
            cond <torch tensor> (B, T2, C) or list of such tensors (from different encoders)
            cond_mask <bool mask> (B, T2): Multiplicative mask where True means we keep the input, False we zero it
                out or list of such tensors (from different encoders) output <torch tensor> (B, T1, C)
            multi_encoder_mapping <list> <int>: None or Same size as n_layers, value indicates which cond input to use
                for this layer

        Returns dict with keys:
            output <torch tensor> (B, T1, C): Output tensor
            attn_probabilities <list>: Attention probabilities of each layer
        """
        if isinstance(cond, list) and len(self.layers) < len(cond):
            raise ValueError(
                f"Insufficient Transformer layers for multiple conditionals. Each layer must cross-attend one conditional."
                f"Found {len(self.layers)} layers for {len(cond)} conditionals."
            )

        if self.use_learnable_pos_emb:
            positions = torch.arange(x.size(1), device=x.device).unsqueeze(0)
            x = x + self.position_embeddings(positions)

        attn_probabilities = []
        x = self.dropout(x)
        for idx, layer in enumerate(self.layers):
            _cond, _cond_mask, _attn_prior = self._get_layer_inputs(
                idx, cond, cond_mask, attn_prior, multi_encoder_mapping
            )
            out_dict = layer(x, x_mask, _cond, _cond_mask, attn_prior=_attn_prior)
            x = out_dict['output']
            attn_probabilities.append(out_dict['attn_probabilities'])

            if max_layer_idx is not None and idx == max_layer_idx:
                break

        x = self.norm_out(x)
        x = self.dropout_out(x)
        return {'output': x, 'attn_probabilities': attn_probabilities}



# Copyright (c) 2021, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import pytest
from omegaconf import DictConfig

from nemo.collections.asr.models.classification_models import EncDecRegressionModel


@pytest.fixture()
def speech_regression_model():
    preprocessor = {'cls': 'nemo.collections.asr.modules.AudioToMelSpectrogramPreprocessor', 'params': dict({})}
    encoder = {
        'cls': 'nemo.collections.asr.modules.ConvASREncoder',
        'params': {
            'feat_in': 64,
            'activation': 'relu',
            'conv_mask': True,
            'jasper': [
                {
                    'filters': 32,
                    'repeat': 1,
                    'kernel': [1],
                    'stride': [1],
                    'dilation': [1],
                    'dropout': 0.0,
                    'residual': False,
                    'separable': True,
                    'se': True,
                    'se_context_size': -1,
                }
            ],
        },
    }

    decoder = {
        'cls': 'nemo.collections.asr.modules.conv_asr.ConvASRDecoderClassification',
        'params': {'feat_in': 32, 'return_logits': True, 'num_classes': 1},
    }

    modelConfig = DictConfig(
        {
            'preprocessor': DictConfig(preprocessor),
            'encoder': DictConfig(encoder),
            'decoder': DictConfig(decoder),
            'labels': None,
            'is_regression_task': True,
        }
    )
    model = EncDecRegressionModel(cfg=modelConfig)
    return model


class TestEncDecRegressionModel:
    @pytest.mark.unit
    def test_constructor(self, speech_regression_model):
        asr_model = speech_regression_model.train()

        conv_cnt = (64 * 32 * 1 + 32) + (64 * 1 * 1 + 32)  # separable kernel + bias + pointwise kernel + bias
        bn_cnt = (4 * 32) * 2  # 2 * moving averages
        dec_cnt = 32 * 1 + 1  # fc + bias

        param_count = conv_cnt + bn_cnt + dec_cnt
        assert asr_model.num_weights == param_count

        # Check to/from config_dict:
        confdict = asr_model.to_config_dict()
        instance2 = EncDecRegressionModel.from_config_dict(confdict)

        assert isinstance(instance2, EncDecRegressionModel)

    @pytest.mark.unit
    def test_transcription(self, speech_regression_model, test_data_dir):

        audio_filenames = ['an22-flrp-b.wav', 'an90-fbbh-b.wav']
        audio_paths = [os.path.join(test_data_dir, "asr", "train", "an4", "wav", fp) for fp in audio_filenames]

        model = speech_regression_model.eval()

        # Test Top 1 classification transcription
        results = model.transcribe(audio_paths, batch_size=2)
        assert len(results) == 2



# Copyright (c) 2023, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# CTC
from nemo.collections.multimodal.speech_cv.models.visual_ctc_bpe_models import VisualEncDecCTCModelBPE
from nemo.collections.multimodal.speech_cv.models.visual_ctc_models import VisualEncDecCTCModel
from nemo.collections.multimodal.speech_cv.models.visual_hybrid_rnnt_ctc_bpe_models import (
    VisualEncDecHybridRNNTCTCBPEModel,
)

# Hybrid CTC/RNN-T
from nemo.collections.multimodal.speech_cv.models.visual_hybrid_rnnt_ctc_models import VisualEncDecHybridRNNTCTCModel
from nemo.collections.multimodal.speech_cv.models.visual_rnnt_bpe_models import VisualEncDecRNNTBPEModel

# RNN-T
from nemo.collections.multimodal.speech_cv.models.visual_rnnt_models import VisualEncDecRNNTModel



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import asyncio
from typing import AsyncGenerator, Optional

import numpy as np
from loguru import logger
from pipecat.frames.frames import (
    CancelFrame,
    EndFrame,
    ErrorFrame,
    Frame,
    StartFrame,
    VADUserStartedSpeakingFrame,
    VADUserStoppedSpeakingFrame,
)
from pipecat.processors.frame_processor import FrameDirection
from pipecat.services.stt_service import STTService
from pipecat.transcriptions.language import Language
from pipecat.utils.time import time_now_iso8601
from pipecat.utils.tracing.service_decorators import traced_stt
from pydantic import BaseModel

from nemo.agents.voice_agent.pipecat.frames.frames import DiarResultFrame
from nemo.agents.voice_agent.pipecat.services.nemo.streaming_diar import DiarizationConfig, NeMoStreamingDiarService


class NeMoDiarInputParams(BaseModel):
    threshold: Optional[float] = (
        0.4  # threshold value used to determine if a speaker exists or not, setting it to a lower value will increase the sensitivity of the diarization model
    )
    language: Optional[Language] = Language.EN_US
    frame_len_in_secs: Optional[float] = 0.08  # 80ms for FastConformer model
    config_path: Optional[str] = None  # path to the Niva ASR config file
    raw_audio_frame_len_in_secs: Optional[float] = 0.016  # 16ms for websocket transport
    buffer_size: Optional[int] = (
        30  # number of audio frames to buffer, 1 frame is 16ms, streaming Sortformer was trained with 6*0.08=0.48s chunks
    )


class NemoDiarService(STTService):
    def __init__(
        self,
        *,
        model: Optional[str] = "",
        device: Optional[str] = "cuda:0",
        sample_rate: Optional[int] = 16000,
        params: Optional[NeMoDiarInputParams] = None,
        use_vad: bool = True,
        audio_passthrough: bool = True,
        backend: Optional[str] = "legacy",
        enabled: bool = True,
        **kwargs,
    ):
        super().__init__(audio_passthrough=audio_passthrough, **kwargs)

        self._enabled = enabled
        self._queue = asyncio.Queue()
        self._response_queue = asyncio.Queue()  # Add response queue
        self._processing_task = None  # Add processing task
        self._response_task = None  # Add response task
        self._device = device
        self._sample_rate = sample_rate
        self._audio_passthrough = audio_passthrough
        params.buffer_size = params.frame_len_in_secs // params.raw_audio_frame_len_in_secs
        self._params = params
        self._model_name = model
        self._use_vad = use_vad
        self._backend = backend
        if not params:
            raise ValueError("params is required")

        self._load_model()

        self._vad_user_speaking = False
        self._audio_buffer = []
        self._current_speaker_id = None
        self._processing_running = False

        if not self._use_vad:
            self._vad_user_speaking = True

    def _load_model(self):
        if not self._enabled or not self._model_name:
            self._model = None
            self._enabled = False
            return

        if self._backend == "legacy":
            cfg = DiarizationConfig()
            cfg.device = self._device
            self._model = NeMoStreamingDiarService(
                cfg, self._model_name, frame_len_in_secs=self._params.frame_len_in_secs, sample_rate=self.sample_rate
            )
        else:
            raise ValueError(f"Invalid backend: {self._backend}")
        logger.info(f"Diarization service initialized on device: {self._device}")

    def can_generate_metrics(self) -> bool:
        """Indicates whether this service can generate metrics.

        Returns:
            bool: True, as this service supports metric generation.
        """
        return True

    async def start(self, frame: StartFrame):
        """Handle service start."""
        await super().start(frame)

        # Initialize the model if not already done
        if not hasattr(self, "_model"):
            self._load_model()

        # Start background processing task
        if not self._processing_task:
            self._processing_task = self.create_task(self._processing_task_handler())

        # Start response handling task
        if not self._response_task:
            self._response_task = self.create_task(self._response_task_handler())

    async def stop(self, frame: EndFrame):
        """Handle service stop."""
        await super().stop(frame)
        await self._stop_tasks()

    async def cancel(self, frame: CancelFrame):
        """Handle service cancellation."""
        await super().cancel(frame)
        await self._stop_tasks()

    async def _stop_tasks(self):
        """Stop background processing tasks."""
        await self._queue.put(None)  # Signal to stop processing
        if self._processing_task:
            await self.cancel_task(self._processing_task)
            self._processing_task = None

        if self._response_task:
            await self.cancel_task(self._response_task)
            self._response_task = None

    def _diarization_processor(self):
        """Background processor that handles diarization calls."""
        try:
            while self._processing_running:
                try:
                    # Get audio from queue - blocking call that will be interrupted by cancellation
                    future = asyncio.run_coroutine_threadsafe(self._queue.get(), self.get_event_loop())
                    audio = future.result()

                    if audio is None:  # Stop signal
                        logger.debug("Received stop signal in background processor")
                        break

                    # Process diarization
                    diar_result = self._model.diarize(audio)

                    # Send result back to async loop
                    asyncio.run_coroutine_threadsafe(self._response_queue.put(diar_result), self.get_event_loop())

                except Exception as e:
                    logger.error(f"Error in background diarization processor: {e}")
                    # Send error back to async loop
                    asyncio.run_coroutine_threadsafe(self._response_queue.put(('error', e)), self.get_event_loop())

        except Exception as e:
            logger.error(f"Background diarization processor fatal error: {e}")
        finally:
            logger.debug("Background diarization processor stopped")

    async def _processing_task_handler(self):
        """Handler for background processing task."""
        try:
            self._processing_running = True
            logger.debug("Starting background processing task")
            await asyncio.to_thread(self._diarization_processor)
        except asyncio.CancelledError:
            logger.debug("Background processing task cancelled")
            self._processing_running = False
            raise
        finally:
            self._processing_running = False

    async def _handle_diarization_result(self, diar_result):
        """Handle diarization result from background processing."""
        try:
            if diar_result is None:
                return
            dominant_speaker_id = self._get_dominant_speaker_id(diar_result)
            # logger.debug(f"Dominant speaker ID: {dominant_speaker_id}")
            if dominant_speaker_id is not None and dominant_speaker_id != self._current_speaker_id:
                self._current_speaker_id = dominant_speaker_id
                logger.debug(f"Pushing DiarResultFrame with speaker {dominant_speaker_id}")
                await self.push_frame(DiarResultFrame(dominant_speaker_id, stream_id="default"))
        except Exception as e:
            logger.error(f"Error handling diarization result: {e}")
            await self.push_frame(
                ErrorFrame(
                    str(e),
                    time_now_iso8601(),
                )
            )

    async def _response_task_handler(self):
        """Handler for processing diarization results."""
        logger.debug("Response task handler started")
        try:
            while True:
                try:
                    result = await self._response_queue.get()

                    if isinstance(result, tuple) and result[0] == 'error':
                        # Handle error from background processing
                        error = result[1]
                        logger.error(f"Error in NeMo Diarization processing: {error}")
                        await self.push_frame(
                            ErrorFrame(
                                str(error),
                                time_now_iso8601(),
                            )
                        )
                    else:
                        # Handle successful diarization result
                        await self._handle_diarization_result(result)

                except Exception as e:
                    logger.error(f"Error in response task handler: {e}")
        except asyncio.CancelledError:
            logger.debug("Response task handler cancelled")
            raise

    async def run_stt(self, audio: bytes) -> AsyncGenerator[Frame, None]:
        """Process audio data and generate transcription frames.

        Args:
            audio: Raw audio bytes to transcribe

        Yields:
            Frame: Transcription frames containing the results
        """
        if self._vad_user_speaking and self._enabled:
            self._audio_buffer.append(audio)
            if len(self._audio_buffer) >= self._params.buffer_size:
                await self.start_ttfb_metrics()
                await self.start_processing_metrics()
                audio = b"".join(self._audio_buffer)
                self._audio_buffer = []
                # Queue audio for background processing
                await self._queue.put(audio)
        yield None

    @traced_stt
    async def _handle_transcription(self, transcript: str, is_final: bool, language: Optional[str] = None):
        """Handle a transcription result.

        Args:
            transcript: The transcribed text
            is_final: Whether this is a final transcription
            language: The language of the transcription
        """
        pass  # Base implementation - can be extended for specific handling needs

    async def set_language(self, language: Language):
        """Update the service's recognition language.

        Args:
            language: New language for recognition
        """
        if self._params:
            self._params.language = language
        else:
            self._params = NeMoDiarInputParams(language=language)

        logger.info(f"Switching STT language to: {language}")

    async def set_model(self, model: str):
        """Update the service's model.

        Args:
            model: New model name/path to use
        """
        await super().set_model(model)
        self._model_name = model
        self._load_model()

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        """Process audio data and generate transcription frames.

        Args:
            audio: Raw audio bytes to transcribe

        Yields:
            Frame: Transcription frames containing the results
        """
        if not self._enabled:
            # if diarization is disabled, just pass the frame through
            await self.push_frame(frame, direction)
            return

        await super().process_frame(frame, direction)
        if isinstance(frame, VADUserStartedSpeakingFrame):
            self._vad_user_speaking = True
            self._audio_buffer = []
            logger.debug("VADUserStartedSpeakingFrame received")
        elif isinstance(frame, VADUserStoppedSpeakingFrame):
            self._vad_user_speaking = False
            logger.debug("VADUserStoppedSpeakingFrame received")
            self._current_speaker_id = None
            self._audio_buffer = []

    def reset(self):
        """Reset the diarization service."""
        self._current_speaker_id = None
        self._audio_buffer = []
        self._vad_user_speaking = False
        self._model.reset_state()

    def _get_dominant_speaker_id(self, spk_pred: np.ndarray):
        spk_pred = (spk_pred > self._params.threshold).astype(int)
        dominant_speaker_id = None
        if spk_pred.sum() > 0:
            # get the dominant speaker id
            # Filter to only keep frames that have any speaker probability > 0.0
            valid_frame_mask = spk_pred.sum(axis=1) > 0

            # Filter diar_result to only keep valid frames
            filtered_diar_result = spk_pred[valid_frame_mask]  # ndarray of shape [num_valid_frames, num_speakers]

            # Get the primary speaker for each valid frame
            primary_spk = np.argmax(filtered_diar_result, axis=1)  # ndarray of shape [num_valid_frames]
            # logger.debug(f"Primary speaker for valid frames: {primary_spk}")

            # count the number of different speakers in the primary speaker sequence
            num_speakers = len(np.unique(primary_spk))
            # logger.debug(f"Number of different speakers: {num_speakers}")

            # If there are multiple speakers, get the dominant one
            if num_speakers > 1:
                # Count occurrences of each speaker
                speaker_counts = np.bincount(primary_spk)
                dominant_speaker_id = np.argmax(speaker_counts)
            else:
                # Only one speaker, return that speaker ID
                dominant_speaker_id = primary_spk[0]
        return dominant_speaker_id



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from abc import ABC, abstractmethod


class PromptTemplate(ABC):
    """
    Base class for prompt templates.
    Derived classes should implement the format and extract methods.
        - format: format the prompt template with the given arguments
        - extract: extract the answer from the response
    """

    @classmethod
    @abstractmethod
    def format(cls, **kwargs) -> str:
        """
        Format the prompt template with the given arguments.
        """
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def extract(cls, response: str) -> str:
        """
        Extract the answer from the response.
        """
        raise NotImplementedError()


class EuroLLMTranslatorPromptTemplate(PromptTemplate):
    """
    Provides a prompt template for the EuroLLM model to perform translation.
    """

    PROMPT_TEMPLATE = (
        "<|im_start|>system\n<|im_end|>\n"
        "<|im_start|>user\n"
        "Translate the following {src_lang} source text to {tgt_lang}. Always output text in the {tgt_lang} language:\n"
        "{src_lang}: {src_text}\n"
        "{tgt_lang}: <|im_end|>\n"
        "<|im_start|>assistant\n"
        "{tgt_text}"
    )

    @classmethod
    def format(
        cls,
        src_lang: str,
        tgt_lang: str,
        src_prefix: str,
        tgt_prefix: str,
        src_context: str = "",
        tgt_context: str = "",
    ) -> str:
        """
        Generate a translation prompt for the EuroLLM model.
        Args:
            src_lang (str): Source language name.
            tgt_lang (str): Target language name.
            src_prefix (str): Source text to translate.
            tgt_prefix (str): Optional target prefix or placeholder for completion.
            src_context (str): Optional source context to start translation from.
            tgt_context (str): Optional target context to start translation from.
        Returns:
            str: Formatted translation prompt.
        """
        src_text = f"{src_context} {src_prefix}"
        tgt_text = f"{tgt_context} {tgt_prefix}"
        src_text = re.sub(r'\s+', ' ', src_text).strip()
        tgt_text = re.sub(r'\s+', ' ', tgt_text).strip()
        return cls.PROMPT_TEMPLATE.format(src_lang=src_lang, tgt_lang=tgt_lang, src_text=src_text, tgt_text=tgt_text)

    @classmethod
    def extract(cls, response: str) -> str:
        """
        Extract the first line of text from a model response.
        Args:
            response (str): The full response from the model.
        Returns:
            str: The text before the first newline.
        """
        return response.split('\n')[0]



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import tempfile

import numpy as np
import pytest
import soundfile as sf
import torch.cuda
from omegaconf import OmegaConf

from nemo.collections.asr.parts.utils.manifest_utils import write_manifest
from nemo.collections.audio.data import audio_to_audio_dataset
from nemo.collections.audio.data.audio_to_audio import (
    ASRAudioProcessor,
    AudioToTargetDataset,
    AudioToTargetWithEmbeddingDataset,
    AudioToTargetWithReferenceDataset,
    _audio_collate_fn,
)
from nemo.collections.audio.data.audio_to_audio_lhotse import (
    LhotseAudioToTargetDataset,
    convert_manifest_nemo_to_lhotse,
)
from nemo.collections.audio.parts.utils.audio import get_segment_start
from nemo.collections.common.data.lhotse import get_lhotse_dataloader_from_config


class TestAudioDatasets:
    @pytest.mark.unit
    @pytest.mark.parametrize('num_channels', [1, 2])
    @pytest.mark.parametrize('num_targets', [1, 3])
    def test_list_to_multichannel(self, num_channels, num_targets):
        """Test conversion of a list of arrays into"""
        random_seed = 42
        num_samples = 1000

        # Generate random signals
        _rng = np.random.default_rng(seed=random_seed)

        # Multi-channel signal
        golden_target = _rng.normal(size=(num_channels * num_targets, num_samples))

        # Create a list of num_targets signals with num_channels channels
        target_list = [golden_target[n * num_channels : (n + 1) * num_channels, :] for n in range(num_targets)]

        # Check the original signal is not modified
        assert (ASRAudioProcessor.list_to_multichannel(golden_target) == golden_target).all()
        # Check the list is converted back to the original signal
        assert (ASRAudioProcessor.list_to_multichannel(target_list) == golden_target).all()

    @pytest.mark.unit
    @pytest.mark.parametrize('num_channels', [1, 2])
    def test_processor_process_audio(self, num_channels):
        """Test signal normalization in process_audio."""
        num_samples = 1000
        num_examples = 30

        signals = ['input_signal', 'target_signal', 'reference_signal']

        for normalization_signal in [None] + signals:
            # Create processor
            processor = ASRAudioProcessor(
                sample_rate=16000, random_offset=False, normalization_signal=normalization_signal
            )

            # Generate random signals
            for n in range(num_examples):
                example = {signal: torch.randn(num_channels, num_samples) for signal in signals}
                processed_example = processor.process_audio(example)

                # Expected scale
                if normalization_signal:
                    scale = 1.0 / (example[normalization_signal].abs().max() + processor.eps)
                else:
                    scale = 1.0

                # Make sure all signals are scaled as expected
                for signal in signals:
                    assert torch.allclose(
                        processed_example[signal], example[signal] * scale
                    ), f'Failed example {n} signal {signal}'

    @pytest.mark.unit
    def test_audio_collate_fn(self):
        """Test `_audio_collate_fn`"""
        batch_size = 16
        random_seed = 42
        atol = 1e-5

        # Generate random signals
        _rng = np.random.default_rng(seed=random_seed)

        signal_to_channels = {
            'input_signal': 2,
            'target_signal': 1,
            'reference_signal': 1,
        }

        signal_to_length = {
            'input_signal': _rng.integers(low=5, high=25, size=batch_size),
            'target_signal': _rng.integers(low=5, high=25, size=batch_size),
            'reference_signal': _rng.integers(low=5, high=25, size=batch_size),
        }

        # Generate batch
        batch = []
        for n in range(batch_size):
            item = dict()
            for signal, num_channels in signal_to_channels.items():
                random_signal = _rng.normal(size=(num_channels, signal_to_length[signal][n]))
                random_signal = np.squeeze(random_signal)  # get rid of channel dimention for single-channel
                item[signal] = torch.tensor(random_signal)
            batch.append(item)

        # Run UUT
        batched = _audio_collate_fn(batch)

        batched_signals = {
            'input_signal': batched[0].cpu().detach().numpy(),
            'target_signal': batched[2].cpu().detach().numpy(),
            'reference_signal': batched[4].cpu().detach().numpy(),
        }

        batched_lengths = {
            'input_signal': batched[1].cpu().detach().numpy(),
            'target_signal': batched[3].cpu().detach().numpy(),
            'reference_signal': batched[5].cpu().detach().numpy(),
        }

        # Check outputs
        for signal, b_signal in batched_signals.items():
            for n in range(batch_size):
                # Check length
                uut_length = batched_lengths[signal][n]
                golden_length = signal_to_length[signal][n]
                assert (
                    uut_length == golden_length
                ), f'Example {n} signal {signal} length mismatch: batched ({uut_length}) != golden ({golden_length})'

                uut_signal = b_signal[n][:uut_length, ...]
                golden_signal = batch[n][signal][:uut_length, ...].cpu().detach().numpy()
                assert np.allclose(
                    uut_signal, golden_signal, atol=atol
                ), f'Example {n} signal {signal} value mismatch.'

    @pytest.mark.unit
    def test_audio_to_target_dataset(self):
        """Test AudioWithTargetDataset in different configurations.

        Test below cover the following:
        1) no constraints
        2) filtering based on signal duration
        3) use with channel selector
        4) use with fixed audio duration and random subsegments
        5) collate a batch of items

        In this use case, each line of the manifest file has the following format:
        ```
        {
            'input_filepath': 'path/to/input.wav',
            'target_filepath': 'path/to/path_to_target.wav',
            'duration': duration_of_input,
        }
        ```
        """
        # Data setup
        random_seed = 42
        sample_rate = 16000
        num_examples = 25
        data_num_channels = {
            'input_signal': 4,
            'target_signal': 2,
        }
        data_min_duration = 2.0
        data_max_duration = 8.0
        data_key = {
            'input_signal': 'input_filepath',
            'target_signal': 'target_filepath',
        }

        # Tolerance
        atol = 1e-6

        # Generate random signals
        _rng = np.random.default_rng(seed=random_seed)

        # Input and target signals have the same duration
        data_duration = np.round(_rng.uniform(low=data_min_duration, high=data_max_duration, size=num_examples), 3)
        data_duration_samples = np.floor(data_duration * sample_rate).astype(int)

        data = dict()
        for signal, num_channels in data_num_channels.items():
            data[signal] = []
            for n in range(num_examples):
                if num_channels == 1:
                    random_signal = _rng.uniform(low=-0.5, high=0.5, size=(data_duration_samples[n]))
                else:
                    random_signal = _rng.uniform(low=-0.5, high=0.5, size=(num_channels, data_duration_samples[n]))
                data[signal].append(random_signal)

        with tempfile.TemporaryDirectory() as test_dir:

            # Build metadata for manifest
            metadata = []

            for n in range(num_examples):

                meta = dict()

                for signal in data:
                    # filenames
                    signal_filename = f'{signal}_{n:02d}.wav'

                    # write audio files
                    sf.write(os.path.join(test_dir, signal_filename), data[signal][n].T, sample_rate, 'float')

                    # update metadata
                    meta[data_key[signal]] = signal_filename

                meta['duration'] = data_duration[n]
                metadata.append(meta)

            # Save manifest
            manifest_filepath = os.path.join(test_dir, 'manifest.json')
            write_manifest(manifest_filepath, metadata)

            # Test 1
            # - No constraints on channels or duration
            dataset = AudioToTargetDataset(
                manifest_filepath=manifest_filepath,
                input_key=data_key['input_signal'],
                target_key=data_key['target_signal'],
                sample_rate=sample_rate,
            )

            # Also test the corresponding factory
            config = {
                'manifest_filepath': manifest_filepath,
                'input_key': data_key['input_signal'],
                'target_key': data_key['target_signal'],
                'sample_rate': sample_rate,
            }
            dataset_factory = audio_to_audio_dataset.get_audio_to_target_dataset(config)

            # Prepare lhotse manifest
            cuts_path = manifest_filepath.replace('.json', '_cuts.jsonl')
            convert_manifest_nemo_to_lhotse(
                input_manifest=manifest_filepath,
                output_manifest=cuts_path,
                input_key=data_key['input_signal'],
                target_key=data_key['target_signal'],
            )

            # Prepare lhotse dataset
            config_lhotse = {
                'cuts_path': cuts_path,
                'use_lhotse': True,
                'sample_rate': sample_rate,
                'batch_size': 1,
            }
            dl_lhotse = get_lhotse_dataloader_from_config(
                OmegaConf.create(config_lhotse), global_rank=0, world_size=1, dataset=LhotseAudioToTargetDataset()
            )
            dataset_lhotse = [item for item in dl_lhotse]

            # Test number of channels
            for signal in data:
                assert data_num_channels[signal] == dataset.num_channels(
                    signal
                ), f'Num channels not correct for signal {signal}'
                assert data_num_channels[signal] == dataset_factory.num_channels(
                    signal
                ), f'Num channels not correct for signal {signal}'

            # Test returned examples
            for n in range(num_examples):
                for signal in data:
                    golden_signal = data[signal][n]

                    for use_lhotse in [False, True]:
                        item_signal = (
                            dataset_lhotse[n][signal].squeeze(0) if use_lhotse else dataset.__getitem__(n)[signal]
                        )
                        item_factory_signal = dataset_factory.__getitem__(n)[signal]

                        assert (
                            item_signal.shape == golden_signal.shape
                        ), f'Test 1, use_lhotse={use_lhotse}: Signal {signal} item shape {item_signal.shape} not matching reference shape {golden_signal.shape}'
                        assert np.allclose(
                            item_signal, golden_signal, atol=atol
                        ), f'Test 1, use_lhotse={use_lhotse}: Failed for example {n}, signal {signal} (random seed {random_seed})'

                        assert np.allclose(
                            item_factory_signal, golden_signal, atol=atol
                        ), f'Test 1, use_lhotse={use_lhotse}: Failed for factory example {n}, signal {signal} (random seed {random_seed})'

            # Test 2
            # - Filtering based on signal duration
            min_duration = 3.5
            max_duration = 7.5

            dataset = AudioToTargetDataset(
                manifest_filepath=manifest_filepath,
                input_key=data_key['input_signal'],
                target_key=data_key['target_signal'],
                min_duration=min_duration,
                max_duration=max_duration,
                sample_rate=sample_rate,
            )

            # Prepare lhotse dataset
            config_lhotse = {
                'cuts_path': cuts_path,
                'use_lhotse': True,
                'min_duration': min_duration,
                'max_duration': max_duration,
                'sample_rate': sample_rate,
                'batch_size': 1,
            }
            dl_lhotse = get_lhotse_dataloader_from_config(
                OmegaConf.create(config_lhotse), global_rank=0, world_size=1, dataset=LhotseAudioToTargetDataset()
            )
            dataset_lhotse = [item for item in dl_lhotse]

            filtered_examples = [n for n, val in enumerate(data_duration) if min_duration <= val <= max_duration]

            for n in range(len(dataset)):
                for use_lhotse in [False, True]:
                    for signal in data:
                        item_signal = (
                            dataset_lhotse[n][signal].squeeze(0) if use_lhotse else dataset.__getitem__(n)[signal]
                        )
                        golden_signal = data[signal][filtered_examples[n]]
                        assert (
                            item_signal.shape == golden_signal.shape
                        ), f'Test 2, use_lhotse={use_lhotse}: Signal {signal} item shape {item_signal.shape} not matching reference shape {golden_signal.shape}'

                        assert np.allclose(
                            item_signal, golden_signal, atol=atol
                        ), f'Test 2, use_lhotse={use_lhotse}: Failed for example {n}, signal {signal} (random seed {random_seed})'

            # Test 3
            # - Use channel selector
            channel_selector = {
                'input_signal': [0, 2],
                'target_signal': 1,
            }

            dataset = AudioToTargetDataset(
                manifest_filepath=manifest_filepath,
                input_key=data_key['input_signal'],
                target_key=data_key['target_signal'],
                input_channel_selector=channel_selector['input_signal'],
                target_channel_selector=channel_selector['target_signal'],
                sample_rate=sample_rate,
            )

            for n in range(len(dataset)):
                item = dataset.__getitem__(n)

                for signal in data:
                    cs = channel_selector[signal]
                    item_signal = item[signal].cpu().detach().numpy()
                    golden_signal = data[signal][n][cs, ...]
                    assert (
                        item_signal.shape == golden_signal.shape
                    ), f'Signal {signal}: item shape {item_signal.shape} not matching reference shape {golden_signal.shape}'
                    assert np.allclose(
                        item_signal, golden_signal, atol=atol
                    ), f'Test 3: Failed for example {n}, signal {signal} (random seed {random_seed})'

            # Test 4
            # - Use fixed duration (random segment selection)
            audio_duration = 4.0
            audio_duration_samples = int(np.floor(audio_duration * sample_rate))

            filtered_examples = [n for n, val in enumerate(data_duration) if val >= audio_duration]

            for random_offset in [True, False]:
                # Test subsegments with the default fixed offset and a random offset

                dataset = AudioToTargetDataset(
                    manifest_filepath=manifest_filepath,
                    input_key=data_key['input_signal'],
                    target_key=data_key['target_signal'],
                    sample_rate=sample_rate,
                    min_duration=audio_duration,
                    audio_duration=audio_duration,
                    random_offset=random_offset,  # random offset when selecting subsegment
                )

                # Prepare lhotse dataset
                config_lhotse = {
                    'cuts_path': cuts_path,
                    'use_lhotse': True,
                    'min_duration': audio_duration,
                    'truncate_duration': audio_duration,
                    'truncate_offset_type': 'random' if random_offset else 'start',
                    'sample_rate': sample_rate,
                    'batch_size': 1,
                }
                dl_lhotse = get_lhotse_dataloader_from_config(
                    OmegaConf.create(config_lhotse), global_rank=0, world_size=1, dataset=LhotseAudioToTargetDataset()
                )
                dataset_lhotse = [item for item in dl_lhotse]

                for n in range(len(dataset)):
                    for use_lhotse in [False, True]:
                        item = dataset_lhotse[n] if use_lhotse else dataset.__getitem__(n)
                        golden_start = golden_end = None
                        for signal in data:
                            item_signal = item[signal].squeeze(0) if use_lhotse else item[signal]
                            full_golden_signal = data[signal][filtered_examples[n]]

                            # Find random segment using correlation on the first channel
                            # of the first signal, and then use it fixed for other signals
                            if golden_start is None:
                                golden_start = get_segment_start(
                                    signal=full_golden_signal[0, :], segment=item_signal[0, :]
                                )
                                if not random_offset:
                                    assert (
                                        golden_start == 0
                                    ), f'Test 4, use_lhotse={use_lhotse}: Expecting the signal to start at 0 when random_offset is False'

                                golden_end = golden_start + audio_duration_samples
                            golden_signal = full_golden_signal[..., golden_start:golden_end]

                            # Test length is correct
                            assert (
                                item_signal.shape[-1] == audio_duration_samples
                            ), f'Test 4, use_lhotse={use_lhotse}: Signal length ({item_signal.shape[-1]}) not matching the expected length ({audio_duration_samples})'

                            assert (
                                item_signal.shape == golden_signal.shape
                            ), f'Test 4, use_lhotse={use_lhotse}: Signal {signal} item shape {item_signal.shape} not matching reference shape {golden_signal.shape}'
                            # Test signal values
                            assert np.allclose(
                                item_signal, golden_signal, atol=atol
                            ), f'Test 4, use_lhotse={use_lhotse}: Failed for example {n}, signal {signal} (random seed {random_seed})'

            # Test 5:
            # - Test collate_fn
            batch_size = 16

            for use_lhotse in [False, True]:
                if use_lhotse:
                    # Get batch from lhotse dataloader
                    config_lhotse['batch_size'] = batch_size
                    dl_lhotse = get_lhotse_dataloader_from_config(
                        OmegaConf.create(config_lhotse),
                        global_rank=0,
                        world_size=1,
                        dataset=LhotseAudioToTargetDataset(),
                    )
                    batched = next(iter(dl_lhotse))
                else:
                    # Get examples from dataset and collate into a batch
                    batch = [dataset.__getitem__(n) for n in range(batch_size)]
                    batched = dataset.collate_fn(batch)

                # Test all shapes and lengths
                for n, signal in enumerate(data.keys()):
                    length = signal.replace('_signal', '_length')

                    if isinstance(batched, dict):
                        signal_shape = batched[signal].shape
                        signal_len = batched[length]
                    else:
                        signal_shape = batched[2 * n].shape
                        signal_len = batched[2 * n + 1]

                    assert signal_shape == (
                        batch_size,
                        data_num_channels[signal],
                        audio_duration_samples,
                    ), f'Test 5, use_lhotse={use_lhotse}: Unexpected signal {signal} shape {signal_shape}'
                    assert (
                        len(signal_len) == batch_size
                    ), f'Test 5, use_lhotse={use_lhotse}: Unexpected length of signal_len ({len(signal_len)})'
                    assert all(
                        signal_len == audio_duration_samples
                    ), f'Test 5, use_lhotse={use_lhotse}: Unexpected signal_len {signal_len}'

    @pytest.mark.unit
    def test_audio_to_target_dataset_with_target_list(self):
        """Test AudioWithTargetDataset when the input manifest has a list
        of audio files in the target key.

        In this use case, each line of the manifest file has the following format:
        ```
        {
            'input_filepath': 'path/to/input.wav',
            'target_filepath': ['path/to/path_to_target_ch0.wav', 'path/to/path_to_target_ch1.wav'],
            'duration': duration_of_input,
        }
        ```
        """
        # Data setup
        random_seed = 42
        sample_rate = 16000
        num_examples = 25
        data_num_channels = {
            'input_signal': 4,
            'target_signal': 2,
        }
        data_min_duration = 2.0
        data_max_duration = 8.0
        data_key = {
            'input_signal': 'input_filepath',
            'target_signal': 'target_filepath',
        }

        # Tolerance
        atol = 1e-6

        # Generate random signals
        _rng = np.random.default_rng(seed=random_seed)

        # Input and target signals have the same duration
        data_duration = np.round(_rng.uniform(low=data_min_duration, high=data_max_duration, size=num_examples), 3)
        data_duration_samples = np.floor(data_duration * sample_rate).astype(int)

        data = dict()
        for signal, num_channels in data_num_channels.items():
            data[signal] = []
            for n in range(num_examples):
                if num_channels == 1:
                    random_signal = _rng.uniform(low=-0.5, high=0.5, size=(data_duration_samples[n]))
                else:
                    random_signal = _rng.uniform(low=-0.5, high=0.5, size=(num_channels, data_duration_samples[n]))
                data[signal].append(random_signal)

        with tempfile.TemporaryDirectory() as test_dir:

            # Build metadata for manifest
            metadata = []

            for n in range(num_examples):

                meta = dict()

                for signal in data:
                    if signal == 'target_signal':
                        # Save targets as individual files
                        signal_filename = []
                        for ch in range(data_num_channels[signal]):
                            # add current filename
                            signal_filename.append(f'{signal}_{n:02d}_ch_{ch}.wav')
                            # write audio file
                            sf.write(
                                os.path.join(test_dir, signal_filename[-1]),
                                data[signal][n][ch, :],
                                sample_rate,
                                'float',
                            )
                    else:
                        # single file
                        signal_filename = f'{signal}_{n:02d}.wav'

                        # write audio files
                        sf.write(os.path.join(test_dir, signal_filename), data[signal][n].T, sample_rate, 'float')

                    # update metadata
                    meta[data_key[signal]] = signal_filename

                meta['duration'] = data_duration[n]
                metadata.append(meta)

            # Save manifest
            manifest_filepath = os.path.join(test_dir, 'manifest.json')
            write_manifest(manifest_filepath, metadata)

            # Test 1
            # - No constraints on channels or duration
            dataset = AudioToTargetDataset(
                manifest_filepath=manifest_filepath,
                input_key=data_key['input_signal'],
                target_key=data_key['target_signal'],
                sample_rate=sample_rate,
            )

            config = {
                'manifest_filepath': manifest_filepath,
                'input_key': data_key['input_signal'],
                'target_key': data_key['target_signal'],
                'sample_rate': sample_rate,
            }
            dataset_factory = audio_to_audio_dataset.get_audio_to_target_dataset(config)

            # Prepare lhotse manifest
            cuts_path = manifest_filepath.replace('.json', '_cuts.jsonl')
            convert_manifest_nemo_to_lhotse(
                input_manifest=manifest_filepath,
                output_manifest=cuts_path,
                input_key=data_key['input_signal'],
                target_key=data_key['target_signal'],
            )

            # Prepare lhotse dataset
            config_lhotse = {
                'cuts_path': cuts_path,
                'use_lhotse': True,
                'sample_rate': sample_rate,
                'batch_size': 1,
            }
            dl_lhotse = get_lhotse_dataloader_from_config(
                OmegaConf.create(config_lhotse), global_rank=0, world_size=1, dataset=LhotseAudioToTargetDataset()
            )
            dataset_lhotse = [item for item in dl_lhotse]

            for n in range(num_examples):
                for use_lhotse in [False, True]:
                    item = dataset_lhotse[n] if use_lhotse else dataset.__getitem__(n)
                    item_factory = dataset_factory.__getitem__(n)
                    for signal in data:
                        item_signal = item[signal].squeeze(0) if use_lhotse else item[signal]
                        golden_signal = data[signal][n]
                        assert (
                            item_signal.shape == golden_signal.shape
                        ), f'Test 1, use_lhotse={use_lhotse}: Signal {signal} item shape {item_signal.shape} not matching reference shape {golden_signal.shape}'
                        assert np.allclose(
                            item_signal, golden_signal, atol=atol
                        ), f'Test 1, use_lhotse={use_lhotse}: Failed for example {n}, signal {signal} (random seed {random_seed})'

                        assert np.allclose(
                            item_factory[signal], golden_signal, atol=atol
                        ), f'Test 1, use_lhotse={use_lhotse}: Failed for factory example {n}, signal {signal} (random seed {random_seed})'

            # Test 2
            # Set target as the first channel of input_filepath and all files listed in target_filepath.
            # In this case, the target will have 3 channels.
            # Note: this is currently not supported by lhotse, so we only test the default dataset here.
            dataset = AudioToTargetDataset(
                manifest_filepath=manifest_filepath,
                input_key=data_key['input_signal'],
                target_key=[data_key['input_signal'], data_key['target_signal']],
                target_channel_selector=0,
                sample_rate=sample_rate,
            )

            for n in range(num_examples):
                item = dataset.__getitem__(n)

                for signal in data:
                    item_signal = item[signal].cpu().detach().numpy()
                    golden_signal = data[signal][n]
                    if signal == 'target_signal':
                        # add the first channel of the input
                        golden_signal = np.concatenate([data['input_signal'][n][0:1, ...], golden_signal], axis=0)
                    assert (
                        item_signal.shape == golden_signal.shape
                    ), f'Signal {signal}: item shape {item_signal.shape} not matching reference shape {golden_signal.shape}'
                    assert np.allclose(
                        item_signal, golden_signal, atol=atol
                    ), f'Test 2: Failed for example {n}, signal {signal} (random seed {random_seed})'

    @pytest.mark.unit
    def test_audio_to_target_dataset_for_inference(self):
        """Test AudioWithTargetDataset when target_key is
        not set, i.e., it is `None`. This is the case, e.g., when
        running inference, and a target is not available.

        In this use case, each line of the manifest file has the following format:
        ```
        {
            'input_filepath': 'path/to/input.wav',
            'duration': duration_of_input,
        }
        ```
        """
        # Data setup
        random_seed = 42
        sample_rate = 16000
        num_examples = 25
        data_num_channels = {
            'input_signal': 4,
        }
        data_min_duration = 2.0
        data_max_duration = 8.0
        data_key = {
            'input_signal': 'input_filepath',
        }

        # Tolerance
        atol = 1e-6

        # Generate random signals
        _rng = np.random.default_rng(seed=random_seed)

        # Input and target signals have the same duration
        data_duration = np.round(_rng.uniform(low=data_min_duration, high=data_max_duration, size=num_examples), 3)
        data_duration_samples = np.floor(data_duration * sample_rate).astype(int)

        data = dict()
        for signal, num_channels in data_num_channels.items():
            data[signal] = []
            for n in range(num_examples):
                if num_channels == 1:
                    random_signal = _rng.uniform(low=-0.5, high=0.5, size=(data_duration_samples[n]))
                else:
                    random_signal = _rng.uniform(low=-0.5, high=0.5, size=(num_channels, data_duration_samples[n]))
                data[signal].append(random_signal)

        with tempfile.TemporaryDirectory() as test_dir:
            # Build metadata for manifest
            metadata = []
            for n in range(num_examples):
                meta = dict()
                for signal in data:
                    # filenames
                    signal_filename = f'{signal}_{n:02d}.wav'
                    # write audio files
                    sf.write(os.path.join(test_dir, signal_filename), data[signal][n].T, sample_rate, 'float')
                    # update metadata
                    meta[data_key[signal]] = signal_filename
                meta['duration'] = data_duration[n]
                metadata.append(meta)

            # Save manifest
            manifest_filepath = os.path.join(test_dir, 'manifest.json')
            write_manifest(manifest_filepath, metadata)

            # Test 1
            # - No constraints on channels or duration
            dataset = AudioToTargetDataset(
                manifest_filepath=manifest_filepath,
                input_key=data_key['input_signal'],
                target_key=None,  # target_signal will be empty
                sample_rate=sample_rate,
            )

            # Also test the corresponding factory
            config = {
                'manifest_filepath': manifest_filepath,
                'input_key': data_key['input_signal'],
                'target_key': None,
                'sample_rate': sample_rate,
            }
            dataset_factory = audio_to_audio_dataset.get_audio_to_target_dataset(config)

            # Prepare lhotse manifest
            cuts_path = manifest_filepath.replace('.json', '_cuts.jsonl')
            convert_manifest_nemo_to_lhotse(
                input_manifest=manifest_filepath,
                output_manifest=cuts_path,
                input_key=data_key['input_signal'],
                target_key=None,
            )

            # Prepare lhotse dataset
            config_lhotse = {
                'cuts_path': cuts_path,
                'use_lhotse': True,
                'sample_rate': sample_rate,
                'batch_size': 1,
            }
            dl_lhotse = get_lhotse_dataloader_from_config(
                OmegaConf.create(config_lhotse), global_rank=0, world_size=1, dataset=LhotseAudioToTargetDataset()
            )
            dataset_lhotse = [item for item in dl_lhotse]

            for n in range(num_examples):

                for label in ['original', 'factory', 'lhotse']:

                    if label == 'original':
                        item = dataset.__getitem__(n)
                    elif label == 'factory':
                        item = dataset_factory.__getitem__(n)
                    elif label == 'lhotse':
                        item = dataset_lhotse[n]
                    else:
                        raise ValueError(f'Unknown label {label}')

                    # Check target is None
                    if 'target_signal' in item:
                        assert item['target_signal'].numel() == 0, f'{label}: target_signal is expected to be empty.'

                    # Check valid signals
                    for signal in data:

                        item_signal = item[signal].squeeze(0) if label == 'lhotse' else item[signal]
                        golden_signal = data[signal][n]
                        assert (
                            item_signal.shape == golden_signal.shape
                        ), f'{label} -- Signal {signal}: item shape {item_signal.shape} not matching reference shape {golden_signal.shape}'
                        assert np.allclose(
                            item_signal, golden_signal, atol=atol
                        ), f'{label} -- Test 1: Failed for example {n}, signal {signal} (random seed {random_seed})'

    @pytest.mark.unit
    def test_audio_to_target_with_reference_dataset(self):
        """Test AudioWithTargetWithReferenceDataset in different configurations.

        1) reference synchronized with input and target
        2) reference not synchronized

        In this use case, each line of the manifest file has the following format:
        ```
        {
            'input_filepath': 'path/to/input.wav',
            'target_filepath': 'path/to/path_to_target.wav',
            'reference_filepath': 'path/to/path_to_reference.wav',
            'duration': duration_of_input,
        }
        ```
        """
        # Data setup
        random_seed = 42
        sample_rate = 16000
        num_examples = 25
        data_num_channels = {
            'input_signal': 4,
            'target_signal': 2,
            'reference_signal': 1,
        }
        data_min_duration = 2.0
        data_max_duration = 8.0
        data_key = {
            'input_signal': 'input_filepath',
            'target_signal': 'target_filepath',
            'reference_signal': 'reference_filepath',
        }

        # Tolerance
        atol = 1e-6

        # Generate random signals
        _rng = np.random.default_rng(seed=random_seed)

        # Input and target signals have the same duration
        data_duration = np.round(_rng.uniform(low=data_min_duration, high=data_max_duration, size=num_examples), 3)
        data_duration_samples = np.floor(data_duration * sample_rate).astype(int)

        data = dict()
        for signal, num_channels in data_num_channels.items():
            data[signal] = []
            for n in range(num_examples):
                if num_channels == 1:
                    random_signal = _rng.uniform(low=-0.5, high=0.5, size=(data_duration_samples[n]))
                else:
                    random_signal = _rng.uniform(low=-0.5, high=0.5, size=(num_channels, data_duration_samples[n]))
                data[signal].append(random_signal)

        with tempfile.TemporaryDirectory() as test_dir:

            # Build metadata for manifest
            metadata = []

            for n in range(num_examples):

                meta = dict()

                for signal in data:
                    # filenames
                    signal_filename = f'{signal}_{n:02d}.wav'

                    # write audio files
                    sf.write(os.path.join(test_dir, signal_filename), data[signal][n].T, sample_rate, 'float')

                    # update metadata
                    meta[data_key[signal]] = signal_filename

                meta['duration'] = data_duration[n]
                metadata.append(meta)

            # Save manifest
            manifest_filepath = os.path.join(test_dir, 'manifest.json')
            write_manifest(manifest_filepath, metadata)

            # Test 1
            # - No constraints on channels or duration
            # - Reference is not synchronized with input and target, so whole reference signal will be loaded
            dataset = AudioToTargetWithReferenceDataset(
                manifest_filepath=manifest_filepath,
                input_key=data_key['input_signal'],
                target_key=data_key['target_signal'],
                reference_key=data_key['reference_signal'],
                reference_is_synchronized=False,
                sample_rate=sample_rate,
            )

            # Also test the corresponding factory
            config = {
                'manifest_filepath': manifest_filepath,
                'input_key': data_key['input_signal'],
                'target_key': data_key['target_signal'],
                'reference_key': data_key['reference_signal'],
                'reference_is_synchronized': False,
                'sample_rate': sample_rate,
            }
            dataset_factory = audio_to_audio_dataset.get_audio_to_target_with_reference_dataset(config)

            for n in range(num_examples):
                item = dataset.__getitem__(n)
                item_factory = dataset_factory.__getitem__(n)

                for signal in data:
                    item_signal = item[signal].cpu().detach().numpy()
                    golden_signal = data[signal][n]
                    assert (
                        item_signal.shape == golden_signal.shape
                    ), f'Signal {signal}: item shape {item_signal.shape} not matching reference shape {golden_signal.shape}'
                    assert np.allclose(
                        item_signal, golden_signal, atol=atol
                    ), f'Test 1: Failed for example {n}, signal {signal} (random seed {random_seed})'

                    item_factory_signal = item_factory[signal].cpu().detach().numpy()
                    assert np.allclose(
                        item_factory_signal, golden_signal, atol=atol
                    ), f'Test 1: Failed for factory example {n}, signal {signal} (random seed {random_seed})'

            # Test 2
            # - Use fixed duration (random segment selection)
            # - Reference is synchronized with input and target, so the same segment of reference signal will be loaded
            audio_duration = 4.0
            audio_duration_samples = int(np.floor(audio_duration * sample_rate))
            dataset = AudioToTargetWithReferenceDataset(
                manifest_filepath=manifest_filepath,
                input_key=data_key['input_signal'],
                target_key=data_key['target_signal'],
                reference_key=data_key['reference_signal'],
                reference_is_synchronized=True,
                sample_rate=sample_rate,
                min_duration=audio_duration,
                audio_duration=audio_duration,
                random_offset=True,
            )

            filtered_examples = [n for n, val in enumerate(data_duration) if val >= audio_duration]

            for n in range(len(dataset)):
                item = dataset.__getitem__(n)

                golden_start = golden_end = None
                for signal in data:
                    item_signal = item[signal].cpu().detach().numpy()
                    full_golden_signal = data[signal][filtered_examples[n]]

                    # Find random segment using correlation on the first channel
                    # of the first signal, and then use it fixed for other signals
                    if golden_start is None:
                        golden_start = get_segment_start(signal=full_golden_signal[0, :], segment=item_signal[0, :])
                        golden_end = golden_start + audio_duration_samples
                    golden_signal = full_golden_signal[..., golden_start:golden_end]

                    # Test length is correct
                    assert (
                        item_signal.shape[-1] == audio_duration_samples
                    ), f'Test 2: Signal {signal} length ({item_signal.shape[-1]}) not matching the expected length ({audio_duration_samples})'

                    # Test signal values
                    assert np.allclose(
                        item_signal, golden_signal, atol=atol
                    ), f'Test 2: Failed for example {n}, signal {signal} (random seed {random_seed})'

            # Test 3
            # - Use fixed duration (random segment selection)
            # - Reference is not synchronized with input and target, so whole reference signal will be loaded
            audio_duration = 4.0
            audio_duration_samples = int(np.floor(audio_duration * sample_rate))
            dataset = AudioToTargetWithReferenceDataset(
                manifest_filepath=manifest_filepath,
                input_key=data_key['input_signal'],
                target_key=data_key['target_signal'],
                reference_key=data_key['reference_signal'],
                reference_is_synchronized=False,
                sample_rate=sample_rate,
                min_duration=audio_duration,
                audio_duration=audio_duration,
                random_offset=True,
            )

            filtered_examples = [n for n, val in enumerate(data_duration) if val >= audio_duration]

            for n in range(len(dataset)):
                item = dataset.__getitem__(n)

                golden_start = golden_end = None
                for signal in data:
                    item_signal = item[signal].cpu().detach().numpy()
                    full_golden_signal = data[signal][filtered_examples[n]]

                    if signal == 'reference_signal':
                        # Complete signal is loaded for reference
                        golden_signal = full_golden_signal
                    else:
                        # Find random segment using correlation on the first channel
                        # of the first signal, and then use it fixed for other signals
                        if golden_start is None:
                            golden_start = get_segment_start(
                                signal=full_golden_signal[0, :], segment=item_signal[0, :]
                            )
                            golden_end = golden_start + audio_duration_samples
                        golden_signal = full_golden_signal[..., golden_start:golden_end]

                        # Test length is correct
                        assert (
                            item_signal.shape[-1] == audio_duration_samples
                        ), f'Test 3: Signal {signal} length ({item_signal.shape[-1]}) not matching the expected length ({audio_duration_samples})'
                    assert (
                        item_signal.shape == golden_signal.shape
                    ), f'Signal {signal}: item shape {item_signal.shape} not matching reference shape {golden_signal.shape}'
                    # Test signal values
                    assert np.allclose(
                        item_signal, golden_signal, atol=atol
                    ), f'Test 3: Failed for example {n}, signal {signal} (random seed {random_seed})'

            # Test 4:
            # - Test collate_fn
            batch_size = 16
            batch = [dataset.__getitem__(n) for n in range(batch_size)]
            _ = dataset.collate_fn(batch)

    @pytest.mark.unit
    def test_audio_to_target_with_embedding_dataset(self):
        """Test AudioWithTargetWithEmbeddingDataset.

        In this use case, each line of the manifest file has the following format:
        ```
        {
            'input_filepath': 'path/to/input.wav',
            'target_filepath': 'path/to/path_to_target.wav',
            'embedding_filepath': 'path/to/path_to_embedding.npy',
            'duration': duration_of_input,
        }
        ```
        """
        # Data setup
        random_seed = 42
        sample_rate = 16000
        num_examples = 25
        data_num_channels = {
            'input_signal': 4,
            'target_signal': 2,
            'embedding_vector': 1,
        }
        data_min_duration = 2.0
        data_max_duration = 8.0
        embedding_length = 64  # 64-dimensional embedding vector
        data_key = {
            'input_signal': 'input_filepath',
            'target_signal': 'target_filepath',
            'embedding_vector': 'embedding_filepath',
        }

        # Tolerance
        atol = 1e-6

        # Generate random signals
        _rng = np.random.default_rng(seed=random_seed)

        # Input and target signals have the same duration
        data_duration = np.round(_rng.uniform(low=data_min_duration, high=data_max_duration, size=num_examples), 3)
        data_duration_samples = np.floor(data_duration * sample_rate).astype(int)

        data = dict()
        for signal, num_channels in data_num_channels.items():
            data[signal] = []
            for n in range(num_examples):
                data_length = embedding_length if signal == 'embedding_vector' else data_duration_samples[n]

                if num_channels == 1:
                    random_signal = _rng.uniform(low=-0.5, high=0.5, size=(data_length))
                else:
                    random_signal = _rng.uniform(low=-0.5, high=0.5, size=(num_channels, data_length))
                data[signal].append(random_signal)

        with tempfile.TemporaryDirectory() as test_dir:

            # Build metadata for manifest
            metadata = []

            for n in range(num_examples):

                meta = dict()

                for signal in data:
                    if signal == 'embedding_vector':
                        signal_filename = f'{signal}_{n:02d}.npy'
                        np.save(os.path.join(test_dir, signal_filename), data[signal][n])

                    else:
                        # filenames
                        signal_filename = f'{signal}_{n:02d}.wav'

                        # write audio files
                        sf.write(os.path.join(test_dir, signal_filename), data[signal][n].T, sample_rate, 'float')

                    # update metadata
                    meta[data_key[signal]] = signal_filename

                meta['duration'] = data_duration[n]
                metadata.append(meta)

            # Save manifest
            manifest_filepath = os.path.join(test_dir, 'manifest.json')
            write_manifest(manifest_filepath, metadata)

            # Test 1
            # - No constraints on channels or duration
            dataset = AudioToTargetWithEmbeddingDataset(
                manifest_filepath=manifest_filepath,
                input_key=data_key['input_signal'],
                target_key=data_key['target_signal'],
                embedding_key=data_key['embedding_vector'],
                sample_rate=sample_rate,
            )

            # Also test the corresponding factory
            config = {
                'manifest_filepath': manifest_filepath,
                'input_key': data_key['input_signal'],
                'target_key': data_key['target_signal'],
                'embedding_key': data_key['embedding_vector'],
                'sample_rate': sample_rate,
            }
            dataset_factory = audio_to_audio_dataset.get_audio_to_target_with_embedding_dataset(config)

            for n in range(num_examples):
                item = dataset.__getitem__(n)
                item_factory = dataset_factory.__getitem__(n)

                for signal in data:
                    item_signal = item[signal].cpu().detach().numpy()
                    golden_signal = data[signal][n]
                    assert (
                        item_signal.shape == golden_signal.shape
                    ), f'Signal {signal}: item shape {item_signal.shape} not matching reference shape {golden_signal.shape}'
                    assert np.allclose(
                        item_signal, golden_signal, atol=atol
                    ), f'Test 1: Failed for example {n}, signal {signal} (random seed {random_seed})'

                    item_factory_signal = item_factory[signal].cpu().detach().numpy()
                    assert np.allclose(
                        item_factory_signal, golden_signal, atol=atol
                    ), f'Test 1: Failed for factory example {n}, signal {signal} (random seed {random_seed})'

            # Test 2:
            # - Test collate_fn
            batch_size = 16
            batch = [dataset.__getitem__(n) for n in range(batch_size)]
            _ = dataset.collate_fn(batch)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from dataclasses import dataclass
from typing import Any, TypeAlias

from nemo.collections.asr.inference.utils.enums import ASROutputGranularity
from nemo.collections.asr.parts.context_biasing.biasing_multi_model import BiasingRequestItemConfig


@dataclass(slots=True)
class ASRRequestOptions:
    """
    Immutable dataclass representing options for a request
    None value means that the option is not set and the default value will be used
    """

    enable_itn: bool = None
    enable_pnc: bool = None
    stop_history_eou: int = None
    asr_output_granularity: ASROutputGranularity | str = None
    language_code: str | None = None
    enable_nmt: bool = None
    source_language: str = None
    target_language: str = None
    biasing_cfg: BiasingRequestItemConfig | None = None

    def __post_init__(self) -> None:
        """
        Post-init hook:
            Converts the asr_output_granularity to ASROutputGranularity if it is a string
        """
        if isinstance(self.asr_output_granularity, str):
            self.asr_output_granularity = ASROutputGranularity.from_str(self.asr_output_granularity)

        if not self.enable_nmt:
            # Forcibly set the source and target languages to None
            self.source_language = None
            self.target_language = None

    def is_word_level_output(self) -> bool:
        """
        Check if the output granularity is word level.
        """
        return self.asr_output_granularity is ASROutputGranularity.WORD

    def is_segment_level_output(self) -> bool:
        """
        Check if the output granularity is segment level.
        """
        return self.asr_output_granularity is ASROutputGranularity.SEGMENT

    @staticmethod
    def _with_default(value: Any, default: Any) -> Any:
        """
        Return the value if it is not None, otherwise return the default value.
        Args:
            value: The value to check.
            default: The default value to return if the value is None.
        Returns:
            The value if it is not None, otherwise return the default value.
        """
        return default if value is None else value

    def augment_with_defaults(
        self,
        default_enable_itn: bool,
        default_enable_pnc: bool,
        default_enable_nmt: bool,
        default_source_language: str,
        default_target_language: str,
        default_stop_history_eou: int,
        default_asr_output_granularity: ASROutputGranularity | str,
        default_language_code: str | None = None,
        biasing_cfg: BiasingRequestItemConfig | None = None,
    ) -> "ASRRequestOptions":
        """
        Fill unset fields with the passed default values.
        Args:
            default_enable_itn (bool): Default enable ITN.
            default_enable_pnc (bool): Default enable PNC.
            default_enable_nmt (bool): Default enable NMT.
            default_source_language (str): Default source language.
            default_target_language (str): Default target language.
            default_stop_history_eou (int): Default stop history EOU.
            default_asr_output_granularity (ASROutputGranularity | str): Default output granularity.
            default_language_code (str | None): Default language code for prompt-enabled models.
            biasing_cfg: Default biasing config or None
        Returns:
            ASRRequestOptions: Augmented options.
        """
        if isinstance(default_asr_output_granularity, str):
            default_asr_output_granularity = ASROutputGranularity.from_str(default_asr_output_granularity)

        enable_itn = self._with_default(self.enable_itn, default_enable_itn)
        enable_pnc = self._with_default(self.enable_pnc, default_enable_pnc)
        enable_nmt = self._with_default(self.enable_nmt, default_enable_nmt)
        if not enable_nmt:
            # Forcibly set the source and target languages to None
            source_language, target_language = None, None
        else:
            source_language = self._with_default(self.source_language, default_source_language)
            target_language = self._with_default(self.target_language, default_target_language)

        stop_history_eou = self._with_default(self.stop_history_eou, default_stop_history_eou)
        granularity = self._with_default(self.asr_output_granularity, default_asr_output_granularity)
        language_code = self._with_default(self.language_code, default_language_code)

        return ASRRequestOptions(
            enable_itn=enable_itn,
            enable_pnc=enable_pnc,
            enable_nmt=enable_nmt,
            source_language=source_language,
            target_language=target_language,
            stop_history_eou=stop_history_eou,
            asr_output_granularity=granularity,
            language_code=language_code,
            biasing_cfg=self.biasing_cfg or biasing_cfg,
        )

    def has_biasing_request(self):
        """Return True if contains non-empty biasing request"""
        return self.biasing_cfg is not None and (not self.biasing_cfg.is_empty())


RequestOptions: TypeAlias = ASRRequestOptions



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from contextlib import contextmanager
from pathlib import Path
from typing import Iterator, Optional, Sequence, Tuple

import lightning.pytorch as pl
import megatron
import pytest
import torch
from lightning.pytorch.utilities.types import EVAL_DATALOADERS, TRAIN_DATALOADERS
from megatron.core import ModelParallelConfig, parallel_state
from torch import Tensor

import nemo.lightning as nl
from nemo.lightning.io.mixin import IOMixin
from nemo.lightning.megatron_parallel import DataT, MegatronLossReduction, ReductionT
from nemo.lightning.pytorch.plugins import MegatronDataSampler


### model environment related utilities
def _reset_megatron_parallel_state():
    """Resets _GLOBAL_NUM_MICROBATCHES_CALCULATOR in megatron which is used in NeMo to initialized model parallel in
    nemo.collections.nlp.modules.common.megatron.megatron_init.initialize_model_parallel_for_nemo
    """  # noqa: D205, D415
    megatron.core.num_microbatches_calculator._GLOBAL_NUM_MICROBATCHES_CALCULATOR = None
    # Clean up any process groups created in testing
    torch.cuda.empty_cache()
    if parallel_state.is_initialized():
        parallel_state.destroy_model_parallel()
    if torch.distributed.is_initialized():
        torch.distributed.destroy_process_group()


@contextmanager
def reset_megatron_parallel_state() -> Iterator[None]:
    """Puts you into a clean parallel state, and again tears it down at the end."""
    try:
        _reset_megatron_parallel_state()
        yield
    finally:
        _reset_megatron_parallel_state()


class RandomDataset(pl.LightningDataModule):
    def __init__(self, size, length):
        super().__init__()
        self.len = length
        self.data = torch.randn(length, size)
        self.data_sampler = MegatronDataSampler(
            seq_len=size,
            micro_batch_size=2,
            global_batch_size=2,
            rampup_batch_size=None,
        )

    def __getitem__(self, index):
        return self.data[index]

    def __len__(self):
        return self.len

    def train_dataloader(self) -> TRAIN_DATALOADERS:
        return torch.utils.data.DataLoader(self.data, batch_size=2)

    def val_dataloader(self) -> EVAL_DATALOADERS:
        return torch.utils.data.DataLoader(self.data, batch_size=2)


class PassThroughLossReduction(MegatronLossReduction):
    """A class used for calculating the loss, and for logging the reduced loss across micro batches."""

    def forward(self, batch: DataT, forward_out: Tensor) -> Tuple[Tensor, ReductionT]:

        return forward_out, forward_out

    def reduce(self, losses_reduced_per_micro_batch: Sequence[ReductionT]) -> Tensor:
        """Works across micro-batches. (data on single gpu).

        Note: This currently only works for logging and this loss will not be used for backpropagation.

        Args:
            losses_reduced_per_micro_batch: a list of the outputs of forward

        Returns:
            A tensor that is the mean of the losses. (used for logging).
        """
        mse_losses = torch.stack([loss for loss in losses_reduced_per_micro_batch])
        return mse_losses.mean()


class ExampleModel(pl.LightningModule, IOMixin):
    def __init__(self, *args, **kwargs):
        super().__init__()

        ## keeps track of number of validation steps
        self.count = torch.zeros((1,))

    def configure_model(self):

        class NestedModel(torch.nn.Module):

            def __init__(self):
                super().__init__()
                self.l1 = torch.nn.modules.Linear(in_features=32, out_features=32)
                self.bn = torch.nn.BatchNorm1d(32)
                self.model_type = "test"
                self.validation_step_outputs = []
                self.vp_stage = None

                class DummyConfig(ModelParallelConfig):
                    calculate_per_token_loss: bool = False
                    fp8: bool = False

                self.config = DummyConfig()

        self.module = NestedModel()

    def forward(self, batch):
        return self.l1(self.bn(batch)).sum()

    def train_dataloader(self):
        dataset = RandomDataset(32, 16)
        return torch.utils.data.DataLoader(dataset, batch_size=2)

    def val_dataloader(self):
        dataset = RandomDataset(32, 16)
        return torch.utils.data.DataLoader(dataset, batch_size=2)

    def test_dataloader(self):
        dataset = RandomDataset(32, 16)
        dl = torch.utils.data.DataLoader(dataset, batch_size=2)
        self._test_names = ['test_{}_'.format(idx) for idx in range(len(dl))]
        return dl

    def training_step(self, batch):
        return self(batch)

    def validation_step(self, batch):
        ## use a dummy validation loss to ensure that loss is decreasing at each step
        ## which guarantees that the -last checkpoints will be symlinks if specified
        self.count += 1
        self.validation_step_outputs.append(-self.count)
        return -self.count

    def test_step(self, batch):
        loss = self(batch)
        self.test_step_outputs.append(loss)
        return loss

    def configure_optimizers(self):
        return torch.optim.SGD(self.parameters(), lr=1e-3)

    def on_validation_epoch_end(self):
        self.log("val_loss", torch.stack(self.validation_step_outputs).mean())
        self.validation_step_outputs.clear()  # free memory

    def set_input_tensor(self, input_tensor: Optional[Tensor]) -> None:
        pass

    def training_loss_reduction(self) -> MegatronLossReduction:  # noqa: D102
        # This is the function that takes batch['loss_mask'] and the logits output by the model and reduces the loss
        return PassThroughLossReduction()

    def validation_loss_reduction(self) -> MegatronLossReduction:  # noqa: D102
        return PassThroughLossReduction()


def setup_test(path, async_save=False, max_epochs=3):
    model = ExampleModel()

    data = RandomDataset(32, 64)

    resume = nl.AutoResume(
        resume_if_exists=True,
        resume_ignore_no_checkpoint=True,
    )

    nemo_logger = nl.NeMoLogger(
        log_dir=path,
        use_datetime_version=False,
    )

    strategy = nl.MegatronStrategy(
        ckpt_async_save=async_save,
        replace_progress_bar=False,
    )

    trainer = nl.Trainer(
        max_epochs=max_epochs,
        devices=1,
        val_check_interval=6,
        log_every_n_steps=4,
        callbacks=nl.ModelCheckpoint(
            monitor="val_loss",
            save_top_k=3,
            save_on_train_epoch_end=True,
            save_context_on_train_end=False,
            filename=f'{{step}}-{{epoch}}-{{val_loss}}-{{consumed_samples}}',
            save_last="link",
        ),
        strategy=strategy,
    )
    nemo_logger.setup(trainer)
    resume.setup(trainer)

    return data, model, trainer


def get_final_checkpoint(checkpoint_dir):
    dist_checkpoints = [d for d in list(checkpoint_dir.glob("*")) if d.is_dir()]
    last_checkpoints = [d for d in dist_checkpoints if d.match("*last")]

    assert len(last_checkpoints) == 1  ## should only have one -last checkpoint
    final_ckpt = last_checkpoints[0]

    top_k_checkpoints = [d for d in dist_checkpoints if d not in last_checkpoints]

    return final_ckpt, top_k_checkpoints


class TestLinkCheckpoint:

    @pytest.mark.unit
    @pytest.mark.run_only_on("GPU")
    def test_link_ckpt(self, tmpdir):
        """Test to ensure that we always keep top_k checkpoints, even after resuming."""

        with reset_megatron_parallel_state():
            tmp_path = tmpdir / "link_ckpt_test"
            data, model, trainer = setup_test(tmp_path, async_save=False)

            trainer.fit(model, data)

            checkpoint_dir = Path(tmp_path / "default" / "checkpoints")
            final_ckpt, top_k_checkpoints = get_final_checkpoint(checkpoint_dir)
            assert os.path.islink(final_ckpt)

            ## make sure we're saving the expected number of checkpoints
            assert len(top_k_checkpoints) == 3

            link = final_ckpt.resolve()
            assert str(final_ckpt).replace("-last", "") == str(link)

    @pytest.mark.unit
    @pytest.mark.run_only_on("GPU")
    def test_link_ckpt_async(self, tmpdir):
        """Test to ensure that we always keep top_k checkpoints, even after resuming."""

        with reset_megatron_parallel_state():
            tmp_path = tmpdir / "async_link_ckpt_test"
            data, model, trainer = setup_test(tmp_path, async_save=True)

            trainer.fit(model, data)

            checkpoint_dir = Path(tmp_path / "default" / "checkpoints")
            final_ckpt, top_k_checkpoints = get_final_checkpoint(checkpoint_dir)
            assert os.path.islink(final_ckpt)
            assert len(top_k_checkpoints) == 3

            link = final_ckpt.resolve()
            assert str(final_ckpt).replace("-last", "") == str(link)

    @pytest.mark.unit
    @pytest.mark.run_only_on("GPU")
    def test_restore_async(self, tmpdir):
        """Test to ensure that we always keep top_k checkpoints, even after resuming."""

        with reset_megatron_parallel_state():
            tmp_path = tmpdir / "async_link_ckpt_test"
            data, model, trainer = setup_test(tmp_path, async_save=True, max_epochs=3)

            trainer.fit(model, data)

            ## reinitialize
            data, model, trainer = setup_test(tmp_path, async_save=True, max_epochs=6)

            trainer.fit(model, data)

            checkpoint_dir = Path(tmp_path / "default" / "checkpoints")
            final_ckpt, top_k_checkpoints = get_final_checkpoint(checkpoint_dir)
            assert os.path.islink(final_ckpt)
            assert len(top_k_checkpoints) == 3

            epoch = str(final_ckpt).split('epoch=')[1][0]
            assert int(epoch) == 5  ## make sure we're running the correct number of epochs



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This is here to import it once, which improves the speed of launch when in debug-mode
from nemo.utils.import_utils import safe_import

safe_import("transformer_engine")

from nemo.collections.llm import peft
from nemo.collections.llm.bert.data import BERTMockDataModule, BERTPreTrainingDataModule, SpecterDataModule
from nemo.collections.llm.bert.model import (
    BertConfig,
    BertEmbeddingLargeConfig,
    BertEmbeddingMiniConfig,
    BertEmbeddingModel,
    BertModel,
    HuggingFaceBertBaseConfig,
    HuggingFaceBertConfig,
    HuggingFaceBertLargeConfig,
    HuggingFaceBertModel,
    MegatronBertBaseConfig,
    MegatronBertConfig,
    MegatronBertLargeConfig,
)
from nemo.collections.llm.gpt.data import (  # noqa: F401
    AlpacaDataModule,
    ChatDataModule,
    CustomReRankerDataModule,
    CustomRetrievalDataModule,
    DollyDataModule,
    FineTuningDataModule,
    HFDatasetDataModule,
    HFDatasetDataModulePacked,
    HFMockDataModule,
    MockDataModule,
    PreTrainingDataModule,
    SpecterReRankerDataModule,
    SquadDataModule,
)
from nemo.collections.llm.gpt.data.api import dolly, hf_dataset, mock, squad
from nemo.collections.llm.gpt.model import (  # noqa: F401
    Baichuan2Config,
    Baichuan2Config7B,
    Baichuan2Model,
    BaseMambaConfig1_3B,
    BaseMambaConfig2_7B,
    BaseMambaConfig130M,
    BaseMambaConfig370M,
    BaseMambaConfig780M,
    ChatGLM2Config6B,
    ChatGLM3Config6B,
    ChatGLMConfig,
    ChatGLMModel,
    CodeGemmaConfig2B,
    CodeGemmaConfig7B,
    CodeLlamaConfig7B,
    CodeLlamaConfig13B,
    CodeLlamaConfig34B,
    CodeLlamaConfig70B,
    DeepSeekModel,
    DeepSeekV2Config,
    DeepSeekV2LiteConfig,
    DeepSeekV3Config,
    Gemma2Config,
    Gemma2Config2B,
    Gemma2Config9B,
    Gemma2Config27B,
    Gemma2Model,
    Gemma3Config1B,
    Gemma3Config4B,
    Gemma3Config12B,
    Gemma3Config27B,
    Gemma3Model,
    GemmaConfig,
    GemmaConfig2B,
    GemmaConfig7B,
    GemmaModel,
    GPTConfig,
    GPTConfig5B,
    GPTConfig7B,
    GPTConfig20B,
    GPTConfig40B,
    GPTConfig126M,
    GPTConfig175B,
    GPTModel,
    GPTOSSConfig,
    GPTOSSConfig20B,
    GPTOSSConfig120B,
    GPTOSSModel,
    Hyena1bConfig,
    Hyena7bARCLongContextConfig,
    Hyena7bConfig,
    Hyena40bARCLongContextConfig,
    Hyena40bConfig,
    HyenaConfig,
    HyenaModel,
    HyenaNV1bConfig,
    HyenaNV7bConfig,
    HyenaNV40bConfig,
    HyenaNVTestConfig,
    HyenaTestConfig,
    Llama2Config7B,
    Llama2Config13B,
    Llama2Config70B,
    Llama3Config8B,
    Llama3Config70B,
    Llama4Config,
    Llama4Experts16Config,
    Llama4Experts128Config,
    Llama31Config8B,
    Llama31Config70B,
    Llama31Config405B,
    Llama31Nemotron70BConfig,
    Llama31NemotronNano8BConfig,
    Llama31NemotronUltra253BConfig,
    Llama32Config1B,
    Llama32Config3B,
    Llama32EmbeddingConfig1B,
    Llama32EmbeddingConfig3B,
    Llama32Reranker1BConfig,
    Llama32Reranker500MConfig,
    Llama33NemotronSuper49BConfig,
    LlamaConfig,
    LlamaEmbeddingModel,
    LlamaModel,
    LlamaNemotronModel,
    MambaModel,
    MaskedTokenLossReduction,
    MistralConfig7B,
    MistralModel,
    MistralNeMoConfig12B,
    MistralSmall3Config24B,
    MixtralConfig,
    MixtralConfig8x3B,
    MixtralConfig8x7B,
    MixtralConfig8x22B,
    MixtralModel,
    Nemotron3Config4B,
    Nemotron3Config8B,
    Nemotron3Config22B,
    Nemotron4Config15B,
    Nemotron4Config340B,
    NemotronConfig,
    NemotronHConfig4B,
    NemotronHConfig8B,
    NemotronHConfig47B,
    NemotronHConfig56B,
    NemotronModel,
    NemotronNano9Bv2,
    NemotronNano12Bv2,
    NVIDIAMambaConfig8B,
    NVIDIAMambaHybridConfig8B,
    Phi3Config,
    Phi3ConfigMini,
    Phi3Model,
    Qwen2Config,
    Qwen2Config1P5B,
    Qwen2Config7B,
    Qwen2Config72B,
    Qwen2Config500M,
    Qwen2Model,
    Qwen3Config,
    Qwen3Config1P7B,
    Qwen3Config4B,
    Qwen3Config8B,
    Qwen3Config14B,
    Qwen3Config30B_A3B,
    Qwen3Config32B,
    Qwen3Config235B_A22B,
    Qwen3Config600M,
    Qwen3Model,
    Qwen25Config1P5B,
    Qwen25Config3B,
    Qwen25Config7B,
    Qwen25Config14B,
    Qwen25Config32B,
    Qwen25Config72B,
    Qwen25Config500M,
    ReRankerModel,
    SSMConfig,
    Starcoder2Config,
    Starcoder2Config3B,
    Starcoder2Config7B,
    Starcoder2Config15B,
    Starcoder2Model,
    StarcoderConfig,
    StarcoderConfig15B,
    StarcoderModel,
    gpt_data_step,
    gpt_forward_step,
)
from nemo.collections.llm.t5.data import FineTuningDataModule as T5FineTuningDataModule
from nemo.collections.llm.t5.data import MockDataModule as T5MockDataModule
from nemo.collections.llm.t5.data import PreTrainingDataModule as T5PreTrainingDataModule
from nemo.collections.llm.t5.data import SquadDataModule as T5SquadDataModule
from nemo.collections.llm.t5.model import (
    T5Config,
    T5Config3B,
    T5Config11B,
    T5Config220M,
    T5Model,
    t5_data_step,
    t5_forward_step,
)

__all__ = [
    "MockDataModule",
    "T5MockDataModule",
    "CustomRetrievalDataModule",
    "CustomReRankerDataModule",
    "SpecterReRankerDataModule",
    "GPTModel",
    "GPTConfig",
    "HyenaTestConfig",
    "Hyena7bConfig",
    "Hyena40bConfig",
    "Hyena7bARCLongContextConfig",
    "Hyena40bARCLongContextConfig",
    "HyenaNVTestConfig",
    "HyenaNV40bConfig",
    "HyenaNV7bConfig",
    "HyenaConfig",
    "HyenaModel",
    "Hyena1bConfig",
    "HyenaNV1bConfig",
    "gpt_data_step",
    "gpt_forward_step",
    "T5Model",
    "T5Config",
    "T5Config220M",
    "T5Config3B",
    "T5Config11B",
    "BertConfig",
    "BertEmbeddingModel",
    "BertModel",
    "BertEmbeddingLargeConfig",
    "BertEmbeddingMiniConfig",
    "t5_data_step",
    "t5_forward_step",
    "MaskedTokenLossReduction",
    "MistralConfig7B",
    "MistralNeMoConfig12B",
    "MistralSmall3Config24B",
    "MistralModel",
    "MixtralConfig",
    "MixtralConfig8x3B",
    "MixtralConfig8x7B",
    "MixtralConfig8x22B",
    "MixtralModel",
    "Starcoder2Config15B",
    "Starcoder2Config",
    "Starcoder2Model",
    "NemotronModel",
    "Nemotron3Config4B",
    "Nemotron3Config8B",
    "Nemotron3Config22B",
    "Nemotron4Config15B",
    "Nemotron4Config340B",
    "NemotronConfig",
    "LlamaEmbeddingModel",
    "Llama32EmbeddingConfig1B",
    "Llama32EmbeddingConfig3B",
    "Phi3Config",
    "Phi3ConfigMini",
    "Phi3Model",
    "SSMConfig",
    "BaseMambaConfig130M",
    "BaseMambaConfig370M",
    "BaseMambaConfig780M",
    "BaseMambaConfig1_3B",
    "BaseMambaConfig2_7B",
    "NVIDIAMambaConfig8B",
    "NVIDIAMambaHybridConfig8B",
    "NemotronHConfig4B",
    "NemotronHConfig8B",
    "NemotronHConfig47B",
    "NemotronHConfig56B",
    "NemotronNano9Bv2",
    "NemotronNano12Bv2",
    "MambaModel",
    "LlamaConfig",
    "Llama2Config7B",
    "Llama2Config13B",
    "Llama2Config70B",
    "Llama3Config8B",
    "Llama3Config70B",
    "Llama31Config8B",
    "Llama31Config70B",
    "Llama31Config405B",
    "Llama32Config1B",
    "Llama32Config3B",
    "Llama4Experts16Config",
    "Llama4Experts128Config",
    "Llama4Config",
    "Llama31NemotronNano8BConfig",
    "Llama31Nemotron70BConfig",
    "Llama33NemotronSuper49BConfig",
    "Llama31NemotronUltra253BConfig",
    "Llama32Reranker500MConfig",
    "Llama32Reranker1BConfig",
    "CodeLlamaConfig7B",
    "CodeLlamaConfig13B",
    "CodeLlamaConfig34B",
    "CodeLlamaConfig70B",
    "LlamaModel",
    "LlamaNemotronModel",
    "GPTOSSConfig",
    "GPTOSSConfig120B",
    "GPTOSSConfig20B",
    "GPTOSSModel",
    "GemmaConfig",
    "GemmaConfig2B",
    "GemmaConfig7B",
    "CodeGemmaConfig2B",
    "CodeGemmaConfig7B",
    "GemmaModel",
    "Gemma2Model",
    "Gemma2Config9B",
    "Gemma2Config",
    "Gemma2Config27B",
    "Gemma2Config2B",
    "Gemma3Model",
    "Gemma3Config1B",
    "Gemma3Config4B",
    "Gemma3Config12B",
    "Gemma3Config27B",
    "Baichuan2Config",
    "Baichuan2Config7B",
    "Baichuan2Model",
    "ChatGLMConfig",
    "ChatGLM2Config6B",
    "ChatGLM3Config6B",
    "ChatGLMModel",
    "Qwen2Model",
    "Qwen2Config7B",
    "Qwen2Config",
    "Qwen2Config500M",
    "Qwen2Config1P5B",
    "Qwen25Config3B",
    "Qwen2Config72B",
    "Qwen25Config500M",
    "Qwen25Config1P5B",
    "Qwen25Config7B",
    "Qwen25Config14B",
    "Qwen25Config32B",
    "Qwen25Config72B",
    "Qwen3Config",
    "Qwen3Config600M",
    "Qwen3Config1P7B",
    "Qwen3Config4B",
    "Qwen3Config8B",
    "Qwen3Config14B",
    "Qwen3Config32B",
    "Qwen3Config30B_A3B",
    "Qwen3Config235B_A22B",
    "Qwen3Model",
    "PreTrainingDataModule",
    "FineTuningDataModule",
    "ChatDataModule",
    "SquadDataModule",
    "T5PreTrainingDataModule",
    "T5FineTuningDataModule",
    "T5SquadDataModule",
    "T5MockDataModule",
    "DeepSeekModel",
    "DeepSeekV2Config",
    "DeepSeekV2LiteConfig",
    "DeepSeekV3Config",
    "HuggingFaceBertBaseConfig",
    "HuggingFaceBertConfig",
    "HuggingFaceBertLargeConfig",
    "HuggingFaceBertModel",
    "MegatronBertBaseConfig",
    "MegatronBertConfig",
    "MegatronBertLargeConfig",
    "BERTMockDataModule",
    "BERTPreTrainingDataModule",
    "SpecterDataModule",
    "DollyDataModule",
    "tokenizer",
    "mock",
    "squad",
    "dolly",
    "peft",
    "hf_dataset",
    "HFMockDataModule",
]


from nemo.utils import logging

try:
    import nemo_run as run  # noqa: F401

    from nemo.collections.llm.api import (  # noqa: F401
        distill,
        export_ckpt,
        finetune,
        generate,
        import_ckpt,
        pretrain,
        prune,
        ptq,
        train,
        validate,
    )
    from nemo.collections.llm.recipes import *  # noqa

    __all__.extend(
        [
            "train",
            "import_ckpt",
            "export_ckpt",
            "pretrain",
            "validate",
            "finetune",
            "generate",
            "prune",
            "ptq",
            "distill",
        ]
    )
except ImportError as error:
    logging.warning(f"Failed to import nemo.collections.llm.[api,recipes]: {error}")

try:
    from nemo.collections.llm.api import deploy  # noqa: F401

    __all__.append("deploy")
except ImportError as error:
    logging.warning(f"The deploy module could not be imported: {error}")

try:
    from nemo.collections.llm.api import evaluate  # noqa: F401

    __all__.append("evaluate")
except ImportError as error:
    logging.warning(f"The evaluate module could not be imported: {error}")


import warnings


warnings.warn(
    "nemo.collections.llm is deprecated and will be removed in a future major NeMo FW container release. "
    "Please refer to the new Megatron-Bridge repository: https://github.com/NVIDIA-NeMo/Megatron-Bridge",
    DeprecationWarning,
    stacklevel=2,
)
